""" Module for variant peptide dict """
from __future__ import annotations
from collections import deque
import itertools
import copy
from typing import Deque, Dict, Iterable, List, Set, Tuple, TYPE_CHECKING
from Bio.Seq import Seq
from Bio import SeqUtils
from moPepGen import aa, get_equivalent, seqvar
from moPepGen.SeqFeature import FeatureLocation
from moPepGen.svgraph.PVGNode import PVGNode
from moPepGen.svgraph.PVGOrf import PVGOrf
from moPepGen.aa import VariantPeptideIdentifier as vpi
from moPepGen.svgraph.SubgraphTree import SubgraphTree
from moPepGen.seqvar import create_variant_w2f


if TYPE_CHECKING:
    from moPepGen.seqvar.VariantRecord import VariantRecord
    from moPepGen.params import CleavageParams
    from moPepGen.circ import CircRNAModel

class VariantPeptideMetadata():
    """ Variant peptide metadata """
    def __init__(self, label:str=None, orf:Tuple[int,int]=None,
            is_pure_circ_rna:bool=False, has_variants:bool=False,
            segments:List[PeptideSegment]=None, check_orf:bool=False):
        """  """
        self.label = label
        self.orf = orf
        self.is_pure_circ_rna = is_pure_circ_rna
        self.has_variants = has_variants
        self.segments = segments or []
        self.check_orf = check_orf

    def get_key(self) -> str:
        """ get key """
        return f"{self.label}|{self.orf}" if self.check_orf else self.label

class PeptideSegment:
    """ Peptide segment """
    def __init__(self, query:FeatureLocation, ref:FeatureLocation, feature_type:str,
            feature_id:str, variant_id:str):
        """ constructor """
        self.query = query
        self.ref = ref
        self.feature_type = feature_type
        self.feature_id = feature_id
        self.variant_id = variant_id

    def __repr__(self):
        """ str """
        return f"<PeptideSegment query={self.query} ref={self.ref}" +\
            f" feature_type={self.feature_type} feature_id={self.feature_id}" +\
            f" variant_id={self.variant_id}>"

    def merge(self, other:PeptideSegment) -> PeptideSegment:
        """ merge """
        query = FeatureLocation(
            start=self.query.start,
            end=other.query.end,
            start_offset=self.query.start_offset,
            end_offset=other.query.end_offset
        )
        if self.ref:
            ref = FeatureLocation(
                start=self.ref.start,
                end=other.ref.end,
                start_offset=self.ref.start_offset,
                end_offset=other.ref.end_offset
            )
        else:
            ref = None
        return PeptideSegment(
            query=query,
            ref=ref,
            feature_type=self.feature_type,
            feature_id=self.feature_id,
            variant_id=self.variant_id
        )

    def is_adjacent(self, other:PeptideSegment) -> bool:
        """ is adjacent """
        if (self.variant_id is None) != (other.variant_id is None) \
                or (self.variant_id is not None \
                    and self.variant_id != other.variant_id) \
                or self.query.end != other.query.start \
                or self.feature_type != other.feature_type \
                or self.feature_id != other.feature_id:
            return False
        if self.ref:
            this_end = self.ref.end * 3 + self.ref.end_offset - self.query.end_offset
            that_start = other.ref.start * 3 + other.ref.start_offset + other.query.start_offset
            if this_end != that_start:
                return False
        return True

    def to_line(self):
        """ to line """
        fields = [
            str(self.query.start),
            str(self.query.end),
            self.feature_type or '.',
            self.feature_id or '.'
        ]
        if self.ref:
            ref_start = self.ref.start * 3 + self.ref.start_offset
            ref_end = self.ref.end * 3 + self.ref.end_offset
            fields += [
                str(ref_start),
                str(ref_end)
            ]
        else:
            fields += ['.', '.']
        fields += [
            str(self.query.start_offset),
            str(self.query.end_offset)
        ]
        if self.variant_id:
            fields.append(self.variant_id)
        else:
            fields.append('.')
        return '\t'.join(fields)

class AnnotatedPeptideLabel:
    """ Annotated peptide label """
    def __init__(self, label:str, segments:List[PeptideSegment]):
        """ constructor """
        self.label = label
        self.segments = segments

    def to_lines(self) -> str:
        """ to line """
        return [f"{self.label}\t{seg.to_line()}" for seg in self.segments]

class MiscleavedNodeSeries():
    """ Helper class when calling for miscleavage peptides. The nodes contained
    by this class can be joined to make a miscleavage peptide. """
    def __init__(self, nodes:List[PVGNode], additional_variants:Set[VariantRecord]):
        """ Constructor """
        self.nodes = nodes
        self.additional_variants = additional_variants
        self._len = None

    def __len__(self) -> int:
        """ Get length of node sequences """
        if self._len is None:
            self._len = sum(len(x.seq.seq) for x in self.nodes)
        return self._len

    def is_too_short(self, param:CleavageParams) -> bool:
        """ Checks whether the sequence is too short """
        return len(self) < param.min_length

    def is_too_long(self, param:CleavageParams) -> bool:
        """ Checks whether the sequence is too long """
        return not(
            len(self) <= param.max_length
            or (
                self.nodes[0].seq.seq.startswith('M')
                and len(self) <= param.max_length + 1
            )
        )

    def has_trailing_selenocysteins(self):
        """ Checks whether the seires of nodes has any selenocystein in any of
        the trailing non C-pop-collapsed node. """
        for node in reversed(self.nodes):
            if node is not self.nodes[-1] and not node.cpop_collapsed:
                return False
            if node.selenocysteines:
                return True
        return False

TypeVariantPeptideMetadataMap = Dict[Seq, Dict[str, VariantPeptideMetadata]]

class MiscleavedNodes():
    """ Helper class for looking for peptides with miscleavages. This class
    defines the collection of nodes in sequence that starts from the same node,
    with up to X allowed miscleavages that makes the miscleaved peptide
    sequences.

    Attributes:
        - `data` (Deque[List[PVGNode]]): The collection of nodes that make all
          miscleaved sequences.
        - `cleavage_params` (CleavageParams): Cleavage related parameters.
        - `orf` (Tuple[int,int]): The ORF start and end locations.
        - `tx_id` (str): Transcript ID.
        - `leading_node` (PVGNode): The start node that the miscleaved peptides
          are called from. This node must present in the PVG graph.
    """
    def __init__(self, data:Deque[MiscleavedNodeSeries],
            cleavage_params:CleavageParams, orfs:List[PVGOrf]=None, tx_id:str=None,
            gene_id:str=None, leading_node:PVGNode=None, subgraphs:SubgraphTree=None,
            is_circ_rna:bool=False):
        """ constructor """
        self.data = data
        self.orfs = orfs or []
        self.tx_id = tx_id
        self.gene_id = gene_id
        self.cleavage_params = cleavage_params
        self.leading_node = leading_node
        self.subgraphs = subgraphs
        self.is_circ_rna = is_circ_rna

    def is_valid_seq(self, seq:Seq, pool:Set[Seq], denylist:Set[str]) -> bool:
        """ Check whether the seq is valid """
        if seq in pool:
            return True
        min_mw = self.cleavage_params.min_mw
        return self.seq_has_valid_size(seq) \
            and seq not in denylist \
            and 'X' not in seq \
            and SeqUtils.molecular_weight(seq, 'protein') >= min_mw

    def seq_has_valid_size(self, seq:Seq=None, size:int=None) -> bool:
        """ Check whether the seq has valid length """
        if seq is None and size is None:
            raise ValueError
        min_length = self.cleavage_params.min_length
        max_length = self.cleavage_params.max_length
        if size is None:
            size = len(seq)

        return min_length <= size <= max_length

    def create_peptide_segments(self, nodes:List[PVGNode]) -> List[PeptideSegment]:
        """ create peptide segments """
        ref_segments:List[PeptideSegment] = []
        var_segments:List[PeptideSegment] = []
        offset = 0
        branch_variants = set()
        for node in nodes:
            for loc in node.seq.locations:
                branch = self.subgraphs[loc.ref.seqname]
                if branch.variant:
                    branch_variants.add(branch.variant.id)
                seg = PeptideSegment(
                    query=loc.query.shift(offset),
                    ref=loc.ref,
                    feature_type=branch.feature_type,
                    feature_id=branch.feature_id,
                    variant_id=branch.variant.id if branch.variant else None
                )
                if ref_segments and ref_segments[-1].is_adjacent(seg):
                    ref_segments[-1] = ref_segments[-1].merge(seg)
                    continue
                ref_segments.append(seg)
            for variant in node.variants:
                if variant.variant.id in branch_variants:
                    continue
                if variant.upstream_cleavage_altering:
                    continue
                if variant.downstream_cleavage_altering:
                    continue
                if len(variant.location) < 1:
                    continue
                seg = PeptideSegment(
                    query=variant.location.shift(offset),
                    ref=None,
                    feature_type=None,
                    feature_id=None,
                    variant_id=variant.variant.id
                )
                if var_segments and var_segments[-1].is_adjacent(seg):
                    var_segments[-1] = var_segments[-1].merge(seg)
                    continue
                var_segments.append(seg)
            offset += len(node.seq.seq)
        segments = ref_segments + var_segments
        segments.sort(key=lambda x:x.query)
        return segments

    def join_miscleaved_peptides(self, pool:TypeVariantPeptideMetadataMap,
            check_variants:bool, additional_variants:List[VariantRecord],
            denylist:Set[str], is_start_codon:bool=False,
            circ_rna:CircRNAModel=None, truncate_sec:bool=False,
            check_external_variants:bool=True, check_orf:bool=False
            ) -> Iterable[Tuple[Seq, VariantPeptideMetadata]]:
        """ join miscleaved peptides and update the peptide pool.

        Args:
            - `check_variants` (bool): When true, only peptides that carries at
              least 1 variant are kept. And when false, all unique peptides are
              reported (e.g. noncoding).
            - `additional_variants` (List[VariantRecord]): Additional variants,
              e.g., start gain, cleavage gain, stop lost.
            - `denylist` (Set[str]): Peptide sequences that should be excluded.
            - `is_start_codon` (bool): Whether the node contains start codon.
        """
        for series in self.data:
            queue = series.nodes
            metadata = VariantPeptideMetadata(check_orf=check_orf)
            seqs_to_join:List[Seq] = []
            size:int = 0
            variants:Dict[str,VariantRecord] = {}
            in_seq_variants:Dict[str,VariantRecord] = {}

            selenocysteines = []

            for i, node in enumerate(queue):
                other = str(node.seq.seq)
                if truncate_sec:
                    if size > 0:
                        selenocysteines += [x.shift(size) for x in node.selenocysteines]
                    else:
                        selenocysteines +=  node.selenocysteines
                seqs_to_join.append(other)
                size += len(other)

                if check_variants:
                    for variant in node.variants:
                        if variant.is_silent:
                            continue
                        if i > 0 and variant.upstream_cleavage_altering:
                            continue
                        if variant.variant.id not in variants:
                            variants[variant.variant.id] = variant.variant
                        if not variant.variant.is_circ_rna():
                            if variant.variant.id not in in_seq_variants:
                                in_seq_variants[variant.variant.id] = variant.variant
                    if i < len(queue) - 1:
                        _node = self.leading_node if i == 0 else node
                        indels = queue[i + 1].upstream_indel_map.get(_node)
                        if indels:
                            for variant in indels:
                                if variant.id not in variants:
                                    variants[variant.id] = variant

            valid_orf = None
            for orf in self.orfs:
                if self.is_circ_rna \
                        and (any(v.is_circ_rna() for v in variants.values()) \
                        or any(v.is_circ_rna() for v in orf.start_gain)):
                    if orf.is_valid_orf_to_misc_nodes(queue, self.subgraphs, circ_rna):
                        if any(n.is_missing_any_variant(in_seq_variants.values()) for n in queue):
                            continue
                        metadata.orf = tuple(orf.orf)
                        valid_orf = orf
                        break
                else:
                    metadata.orf = tuple(orf.orf)
                    valid_orf = orf
                    break

            if valid_orf is None:
                continue

            cleavage_gain_down = queue[-1].get_cleavage_gain_from_downstream()
            for v in cleavage_gain_down:
                if v not in variants:
                    variants[v.id] = v

            if check_variants:
                for v in valid_orf.start_gain:
                    if v.id not in variants:
                        variants[v.id] = v
                for v in additional_variants:
                    if v.id not in variants:
                        variants[v.id] = v
                for v in valid_orf.cleavage_gain:
                    if v.id not in variants:
                        variants[v.id] = v
                for v in series.additional_variants:
                    if v.id not in variants:
                        variants[v.id] = v

            if self.is_circ_rna \
                    and any(v.is_circ_rna() for v in variants.values()) \
                    and queue[-1].any_unaccounted_downstream_cleavage_or_stop_altering(
                        {x for x in variants.values() if not x.is_circ_rna()}):
                continue

            if size == 0:
                continue

            if check_variants:
                if check_external_variants:
                    if not variants:
                        continue
                else:
                    if not (variants or selenocysteines):
                        continue

            if not selenocysteines \
                    and not self.seq_has_valid_size(size=size) \
                    and not (seqs_to_join[0].startswith('M')
                                and self.seq_has_valid_size(size=size-1)):
                continue

            seq = Seq(''.join(seqs_to_join))
            is_in_denylist = seq in denylist and (not is_start_codon or seq[1:] in denylist)
            if not seq in pool and is_in_denylist:
                continue

            metadata.is_pure_circ_rna = self.is_circ_rna \
                and len(variants) == 1 \
                and list(variants.values())[0].is_circ_rna()

            seqs = self.translational_modification(seq, metadata, denylist,
                variants.values(), is_start_codon, selenocysteines,
                check_variants, check_external_variants, pool, queue
            )
            for seq, metadata in seqs:
                yield seq, metadata


    def translational_modification(self, seq:Seq, metadata:VariantPeptideMetadata,
            denylist:Set[str], variants:Set[VariantRecord], is_start_codon:bool,
            selenocysteines:List[seqvar.VariantRecordWithCoordinate],
            check_variants:bool, check_external_variants:bool, pool:Set[Seq],
            nodes:List[PVGNode]
            ) -> Iterable[Tuple[Seq,VariantPeptideMetadata]]:
        """ Apply any modification that could happen during translation. The
        kinds of modifications that could happen are:
        1. Leading Methionine truncation.
        2. Selenocysteine truncation.
        """
        if variants or not check_variants:
            is_valid = self.is_valid_seq(seq, pool, denylist)

            is_valid_start =  is_start_codon and seq.startswith('M') and\
                self.is_valid_seq(seq[1:], pool, denylist)

            if is_valid or is_valid_start:
                cur_metadata = copy.copy(metadata)
                label = vpi.create_variant_peptide_id(
                    transcript_id=self.tx_id, variants=variants, orf_id=None,
                    gene_id=self.gene_id
                )
                cur_metadata.label = label
                cur_metadata.has_variants = bool(variants)

                if is_valid:
                    cur_metadata_2 = copy.copy(cur_metadata)
                    cur_metadata_2.segments = self.create_peptide_segments(nodes)
                    yield seq, cur_metadata_2

                if is_valid_start:
                    cur_seq=seq[1:]
                    cur_nodes = list(nodes)
                    cur_nodes[0] = cur_nodes[0].copy()
                    cur_nodes[0].truncate_left(1)
                    cur_metadata.segments = self.create_peptide_segments(cur_nodes)
                    yield cur_seq, cur_metadata

        # Selenocysteine termination
        for sec in selenocysteines:
            seq_mod = seq[:sec.location.start]
            is_valid = self.is_valid_seq(seq_mod, pool, denylist)
            is_valid_start = is_start_codon and seq_mod.startswith('M') and\
                self.is_valid_seq(seq_mod[1:], pool, denylist)

            if is_valid or is_valid_start:
                cur_metadata = copy.copy(metadata)
                cur_variants = [v for v in variants if v.location.end
                    <= sec.variant.location.start]
                if check_variants and check_external_variants and not cur_variants:
                    continue
                cur_variants.append(sec.variant)
                label = vpi.create_variant_peptide_id(
                    transcript_id=self.tx_id, variants=cur_variants, orf_id=None,
                    gene_id=self.gene_id
                )
                cur_metadata.label = label
                cur_metadata.has_variants = bool(cur_variants)

                if is_valid:
                    cur_metadata_2 = copy.copy(cur_metadata)
                    cur_nodes = []
                    cut_offset = sec.location.start
                    for node in nodes:
                        if cut_offset == 0:
                            cur_nodes.append(node)
                            continue
                        if len(node.seq.seq) > cut_offset:
                            node = node.copy()
                            node.truncate_left(cut_offset)
                            cur_nodes.append(node)
                            cut_offset = 0
                        else:
                            cut_offset = max(0, cut_offset - len(node.seq.seq))
                            continue
                    cur_metadata_2.segments = self.create_peptide_segments(cur_nodes)
                    yield seq_mod, cur_metadata_2

                if is_valid_start:
                    cur_seq=seq_mod[1:]
                    cur_nodes[0] = cur_nodes[0].copy()
                    cur_nodes[0].truncate_left(1)
                    cur_metadata.segments = self.create_peptide_segments(cur_nodes)
                    yield cur_seq, cur_metadata

    @staticmethod
    def any_overlaps_and_all_missing_variant(nodes:Iterable[PVGNode], variant:VariantRecord):
        """ Whether any node overlaps with a given variant and all nodes are
        missing it. Should be only used for circRNA. """
        any_overlap = False
        all_missing = True
        for node in nodes:
            locs = []
            if node.seq.locations:
                for loc in node.seq.locations:
                    tx_loc = FeatureLocation(
                        start=int(loc.ref.start) * 3 + node.reading_frame_index,
                        end=int(loc.ref.end) * 3 + node.reading_frame_index
                    )
                    locs.append(tx_loc)
            if node.variants:
                locs += [x.variant.location for x in node.variants
                    if not x.variant.is_circ_rna()]
            node_variants = {x.variant for x in node.variants}
            if not any(loc.overlaps(variant.location) for loc in locs):
                continue
            any_overlap = True
            for loc in locs:
                if loc.overlaps(variant.location) \
                        and any(variant == v for v in node_variants):
                    all_missing = False
        return any_overlap and all_missing


def update_peptide_pool(seq:aa.AminoAcidSeqRecord,
        peptide_pool:Set[aa.AminoAcidSeqDict],
        label_counter:Set[aa.AminoAcidSeqRecord], label:str,
        update_label:bool=True) -> None:
    """ Add a peptide sequence to a peptide pool if not already exists,
    otherwise update the same peptide's description """
    if 'X' in seq.seq:
        return
    if update_label:
        if label not in label_counter:
            label_counter[label] = 0
        label_counter[label] += 1
        seq.description = label + '|' + str(label_counter[label])
        seq.id = seq.description
        seq.name = seq.description
        same_peptide = get_equivalent(peptide_pool, seq)
        if same_peptide:
            same_peptide:aa.AminoAcidSeqRecord
            same_peptide.description += ' ' + seq.description
            same_peptide.id = same_peptide.description
            same_peptide.name = same_peptide.description
        else:
            peptide_pool.add(seq)
    else:
        if label not in label_counter:
            label_counter[label] = 0
        label_counter[label] += 1
        seq.description = label + '|'\
            + str(label_counter[label])
        peptide_pool.add(seq)

class VariantPeptideDict():
    """ Variant peptide pool as dict.

    Attributes:
        - `tx_id` (str): Transcript ID.
        - `peptides` (Dict[AminoAcidSeqRecord, Set[VariantPeptideMetadata]]):
          The peptide data pool, with keys being the AminoAcidRecord, and
          values being set of VariantPeptdieMetadata (variants and ORF).
        - `labels` (Dict[str,int]): Label counter, as a dict with key being the
          variant peptide label, and values being the number of occurrence
          of this label.
        - `global_variant` (VariantRecord): A variant that should be applied to
          every node. This is useful for PVG derived from a circRNA, and this
          argument is the circRNA variant.
    """
    def __init__(self, tx_id:str, peptides:TypeVariantPeptideMetadataMap=None,
            seqs:Set[Seq]=None, labels:Dict[str,int]=None,
            global_variant:VariantRecord=None, gene_id:str=None,
            truncate_sec:bool=False, w2f:bool=False, check_external_variants:bool=True,
            cleavage_params:CleavageParams=None, check_orf:bool=False):
        """ constructor """
        self.tx_id = tx_id
        self.peptides = peptides or {}
        self.seqs = seqs or set()
        self.labels = labels or {}
        self.global_variant = global_variant
        self.gene_id = gene_id
        self.truncate_sec = truncate_sec
        self.w2f = w2f
        self.check_external_variants = check_external_variants
        self.cleavage_params = cleavage_params
        self.check_orf = check_orf

    def find_miscleaved_nodes(self, node:PVGNode, orfs:List[PVGOrf],
            cleavage_params:CleavageParams, tx_id:str, gene_id:str,
            leading_node:PVGNode, subgraphs:SubgraphTree, is_circ_rna:bool,
            backsplicing_only:bool
            ) -> MiscleavedNodes:
        """ Find all miscleaved nodes.

        node vs leading_node:
            - When calling this function from a node within a PVG graph, a copy
              of the node should be first created. This is because if the ORF
              start site is located in the node, the peptide sequence should be
              called form the ORF start site, rather than the beginning of the
              sequence. So then a copy should be created from this node and
              only keep the sequence after the ORF start.
            - The `leading_node` must tbe the original copy of the node within
              the graph, that the `node` is copied from. This is used to
              retrieve INDELs from the downstream node.

        Args:
            - `node` (PVGNode): The node that miscleaved peptide sequences
              starts from it should be called.
            - `orfs` (List[PVGPOrf]): The ORF start and end locations.
            - `cleavage_params` (CleavageParams): Cleavage related parameters.
            - `tx_id` (str): Transcript ID.
            - `leading_node` (PVGNode): The start node that the miscleaved
              peptides are called from. This node must present in the PVG graph.
        """
        if not orfs:
            raise ValueError('ORFs are empty')
        queue = deque([[node]])
        nodes = MiscleavedNodes(
            data=deque([]),
            cleavage_params=cleavage_params,
            orfs=orfs,
            tx_id=tx_id,
            gene_id=gene_id,
            leading_node=leading_node,
            subgraphs=subgraphs,
            is_circ_rna=is_circ_rna
        )

        if not (node.cpop_collapsed or node.truncated) and \
                (not backsplicing_only or len(node.get_subgraph_id_set()) > 1):
            additional_variants = leading_node.get_downstream_stop_altering_variants()
            series = MiscleavedNodeSeries([node], additional_variants)
            if series.is_too_long(self.cleavage_params) and not node.selenocysteines:
                return nodes
            if not series.is_too_short(self.cleavage_params):
                nodes.data.append(series)

        while queue:
            cur_batch = queue.pop()
            cur_node = cur_batch[-1]
            # Turn it into a doct of id to variant
            batch_vars:Dict[str, VariantRecord] = {}

            for _node in cur_batch:
                for var in _node.variants:
                    if var.variant.id not in batch_vars:
                        batch_vars[var.variant.id] = var.variant

            n_cleavages = len([x for x in cur_batch if not x.cpop_collapsed]) - 1

            if n_cleavages >= cleavage_params.miscleavage:
                continue

            # This is done to reduce the complexity when the node has too
            # many out nodes, and its out nodes also have too many out nodes.
            if cleavage_params.additional_variants_per_misc == -1:
                allowed_n_vars = float('Inf')
            else:
                allowed_n_vars = cleavage_params.max_variants_per_node
                allowed_n_vars += (n_cleavages + 1) * cleavage_params.additional_variants_per_misc

            for _node in cur_node.out_nodes:
                if is_circ_rna and _node.is_hybrid_node(subgraphs):
                    continue
                if _node.truncated:
                    continue
                new_batch = copy.copy(cur_batch)

                is_stop = len(_node.seq.seq) == 1 and _node.seq.seq.startswith('*')
                if is_stop:
                    continue

                new_batch.append(_node)

                if backsplicing_only:
                    # Skip peptides not spaning the backsplicing site
                    subgraph_ids = set().union(*[n.get_subgraph_id_set() for n in cur_batch])
                    if len(subgraph_ids) == 1:
                        continue

                additional_variants = _node.get_downstream_stop_altering_variants()

                cur_vars = set(batch_vars.keys())
                for var in _node.variants:
                    cur_vars.add(var.variant.id)
                cur_vars.update({v.id for v in additional_variants})
                if len(cur_vars) > allowed_n_vars:
                    continue

                if not _node.cpop_collapsed:
                    series = MiscleavedNodeSeries(copy.copy(new_batch), additional_variants)
                    if series.is_too_long(self.cleavage_params) \
                            and not series.has_trailing_selenocysteins():
                        continue
                    if not series.is_too_short(self.cleavage_params):
                        nodes.data.append(series)
                    if n_cleavages + 1 == cleavage_params.miscleavage:
                        continue
                    if n_cleavages + 1 > cleavage_params.miscleavage:
                        raise ValueError('Something just went wrong')
                queue.append(new_batch)
        return nodes

    def add_miscleaved_sequences(self, node:PVGNode, orfs:List[PVGOrf],
            cleavage_params:CleavageParams,
            check_variants:bool, is_start_codon:bool,
            additional_variants:List[VariantRecord], denylist:Set[str],
            leading_node:PVGNode=None, subgraphs:SubgraphTree=None,
            circ_rna:CircRNAModel=None, backsplicing_only:bool=False):
        """ Add amino acid sequences starting from the given node, with number
        of miscleavages no more than a given number. The sequences being added
        are the sequence of the current node, and plus n of downstream nodes,
        with n ranges from 1 to miscleavages.

        Args:
        - `node` (PVGNode): The start node.
        - `orfs` (List[int, int]): The start and end position of the ORF.
        - `cleavage_params` (CleavageParams): Cleavage related parameters.
        - `check_variants` (bool): Whether to check for variants.
        - `is_start_codon` (bool): Whether the peptide starts with the start
            codon (M).
        - `additional_variants` (List[VariantRecord]): Additional variants,
            e.g., start gain, cleavage gain, stop lost.
        - `denylist` (Set[str]): Peptide sequences that should be excluded.
        - `leading_node` (PVGNode): The start node that the miscleaved
            peptides are called from. This node must be present in the PVG graph.
        - `subgraphs` (SubgraphTree): The subgraph tree that holds how subgraphs
            are incorporated into the main graph.
        - `circ_rna` (CircRNAModel): The circRNA molecular model from which variant
            peptides are being called.
        - `backsplicing_only` (bool): Whether to only output variant peptides
            spanning the backsplicing site.
        """
        if leading_node is None:
            leading_node = node
        miscleaved_nodes = self.find_miscleaved_nodes(
            node=node, orfs=orfs, cleavage_params=cleavage_params,
            tx_id=self.tx_id, gene_id=self.gene_id, leading_node=leading_node,
            subgraphs=subgraphs, is_circ_rna=circ_rna is not None,
            backsplicing_only=backsplicing_only
        )
        if self.global_variant and self.global_variant not in additional_variants:
            additional_variants.append(self.global_variant)

        it = miscleaved_nodes.join_miscleaved_peptides(
            pool=self.peptides,
            check_variants=check_variants,
            additional_variants=additional_variants,
            denylist=denylist,
            is_start_codon=is_start_codon,
            circ_rna=circ_rna,
            truncate_sec=self.truncate_sec,
            check_external_variants=self.check_external_variants,
            check_orf=self.check_orf
        )
        for seq, metadata in it:
            if 'X' in seq:
                continue
            if '*' in seq:
                raise ValueError('Invalid amino acid symbol found in the sequence.')
            val = self.peptides.setdefault(seq, {})
            key = metadata.get_key()
            if key not in val:
                val[key] = metadata
            self.seqs.add(seq)

    def find_codon_reassignments(self, seq:Seq, w2f:bool=False) -> List[VariantRecord]:
        """ Find potential codon reassignments from the given sequence. """
        variants = []
        if w2f:
            i = 0
            while i > -1:
                i = seq.find('W', start=i)
                if i > -1:
                    w2f_var = create_variant_w2f(self.tx_id, i)
                    variants.append(w2f_var)
                    i += 1
                    if i >= len(seq):
                        break
        return variants

    def is_valid_seq(self, seq:Seq, denylist:Set[str]) -> bool:
        """ Check whether the seq is valid """
        if seq in self.seqs:
            return True
        min_mw = self.cleavage_params.min_mw
        return self.seq_has_valid_size(seq) \
            and seq not in denylist \
            and 'X' not in seq \
            and SeqUtils.molecular_weight(seq, 'protein') >= min_mw

    def seq_has_valid_size(self, seq:Seq) -> bool:
        """ Check whether the seq has valid length """
        min_length = self.cleavage_params.min_length
        max_length = self.cleavage_params.max_length

        return min_length <= len(seq) <= max_length

    def translational_modification(self, w2f:bool, denylist:Set[str]):
        """ Sequence level translational modification, e.g., W2F reassignment """
        for seq in copy.copy(self.peptides):
            reassignments:List[seqvar.VariantRecord] = []
            reassignments += self.find_codon_reassignments(seq, w2f)
            if not reassignments:
                continue

            # W > F codon reassignments
            for k in range(1, len(reassignments) + 1):
                for comb in itertools.combinations(reassignments, k):
                    seq_mod = copy.copy(seq)
                    for v in comb:
                        seq_mod_new = seq_mod[:v.location.start] + v.alt
                        if v.location.end < len(seq):
                            seq_mod_new += seq_mod[v.location.end:]
                        seq_mod = seq_mod_new

                    if not self.is_valid_seq(seq_mod, denylist):
                        continue
                    for metadata in self.peptides[seq].values():
                        cur_metadata = copy.copy(metadata)
                        cur_metadata.label += '|' + '|'.join(v.id for v in comb)
                        cur_metadata.has_variants = True

                        val = self.peptides.setdefault(seq_mod, {})
                        key = cur_metadata.get_key()
                        if key not in val:
                            val[key] = cur_metadata
                        self.seqs.add(seq_mod)

    def get_peptide_sequences(self, keep_all_occurrence:bool=True,
            orf_id_map:Dict[Tuple[int,int],str]=None, check_variants:bool=True
            ) -> Dict[Seq,List[AnnotatedPeptideLabel]]:
        """ Get the peptide sequence and check for miscleavages.

        Args:
            - `keep_all_occurrence` (bool): Whether to keep all occurance of
              the peptide within the graph/transcript.
            - `orf_id_map` (Dict[int, str]): An ID map of ORF IDs from ORF
              start position.
        """
        peptide_segments: Dict[Seq, List[AnnotatedPeptideLabel]] = {}

        for seq in self.peptides:
            if '*' in seq:
                raise ValueError('Invalid amino acid symbol found in the sequence.')
            labels = []
            metadatas = list(self.peptides[seq].values())
            metadatas.sort(key=lambda x: x.orf[0])
            unique_labels = set()
            had_pure_circ_rna = False
            for metadata in metadatas:
                if check_variants and not metadata.has_variants:
                    continue
                if metadata.is_pure_circ_rna:
                    if had_pure_circ_rna:
                        continue
                    had_pure_circ_rna = True
                orf_id = None
                if orf_id_map:
                    orf_id = orf_id_map[metadata.orf]

                label = metadata.label
                if orf_id:
                    label += f"|{orf_id}"

                if label in unique_labels:
                    continue
                unique_labels.add(label)

                if label in self.labels:
                    self.labels[label] += 1
                else:
                    self.labels[label] = 1
                label += f"|{self.labels[label]}"
                labels.append(label)
                seq_anno = AnnotatedPeptideLabel(label, metadata.segments)
                if seq in peptide_segments:
                    peptide_segments[seq].append(seq_anno)
                else:
                    peptide_segments[seq] = [seq_anno]

                if not keep_all_occurrence:
                    break

            if not labels:
                continue

        return peptide_segments



""" Extract records from GVF files """
import argparse
from pathlib import Path
from typing import List
from moPepGen import seqvar
from moPepGen import circ
from moPepGen.cli.common import print_help_if_missing_args, add_args_debug_level
from moPepGen.seqvar.GVFMetadata import GVFMetadata
from moPepGen.circ import CircRNAModel


# pylint: disable=W0212
def parse_args(subparsers:argparse._SubParsersAction):
    """ add args """
    parser:argparse.ArgumentParser = subparsers.add_parser(
        name='extractGVF',
        help='Extract records from GVF files'
    )
    parser.add_argument(
        '-i', '--input-gvf',
        type=Path,
        help='Input GVF files',
        nargs='*',
        metavar='<files>',
        required=True
    )
    parser.add_argument(
        '-o', '--output-dir',
        type=Path,
        help='Output directory',
        metavar='<file>',
        required=True
    )
    parser.add_argument(
        '--gene-list',
        type=str,
        help='Gene list to filter.',
        nargs='*',
        metavar='<values>',
        default=[]
    )
    parser.add_argument(
        '--tx-list',
        type=str,
        help='Transcript list to filter.',
        nargs='*',
        metavar='<values>',
        default=[]
    )
    parser.set_defaults(func=main)
    add_args_debug_level(parser)
    print_help_if_missing_args(parser)
    return parser

def is_wanted_variant(variant:seqvar.VariantRecord, tx_list:List[str],
        gene_list:List[str]) -> bool:
    """ check if the variant record is wanted """
    return variant.location.seqname in gene_list or \
        variant.attrs['TRANSCRIPT_ID'] in tx_list

def is_wanted_circ_rna(variant:CircRNAModel, tx_list:List[str],
        gene_list:List[str]) -> bool:
    """ check if the circRNA record is wanted """
    return variant.gene_id in gene_list or variant.transcript_id in tx_list

def main(args:argparse.Namespace):
    """ Extract records from GVF files """
    gvf_files:List[Path] = args.input_gvf
    gene_list:List[str] = args.gene_list
    tx_list:List[str] = args.tx_list
    output_dir:Path = args.output_dir

    for gvf_file in gvf_files:
        with open(gvf_file, 'rt') as handle:
            metadata = GVFMetadata.parse(handle)
            output_file = output_dir/f"{metadata.source}.gvf"
            if metadata.is_circ_rna():
                records:List[CircRNAModel] = []
                for record in circ.io.parse(handle):
                    if is_wanted_circ_rna(record, tx_list, gene_list):
                        records.append(record)
                with open(output_file, 'wt') as out_handle:
                    circ.io.write(records, metadata, out_handle)
            else:
                records:List[seqvar.VariantRecord] = []
                for record in seqvar.io.parse(handle):
                    if is_wanted_variant(record, tx_list, gene_list):
                        records.append(record)
                seqvar.io.write(records, output_file, metadata)



""" This module defines the class logic for the GTF annotations.
"""
from __future__ import annotations
import copy
from typing import List, Dict, Tuple, TYPE_CHECKING, IO, Union
from moPepGen.SeqFeature import FeatureLocation, SeqFeature
from moPepGen import seqvar, err
from moPepGen.version import MetaVersion
from . import GtfIO
from .TranscriptAnnotationModel import TranscriptAnnotationModel, GTF_FEATURE_TYPES
from .GeneAnnotationModel import GeneAnnotationModel
from .GTFSeqFeature import GTFSeqFeature
from .GTFSourceInferrer import GTFSourceInferrer


if TYPE_CHECKING:
    from moPepGen.aa import AminoAcidSeqDict

class GenomicAnnotation():
    """ This defines the annotation of genes and transcripts of the genome,
    reading from a GTF file.

    Attributes:
        genes (Dict[str,GeneAnnotationModel]): Keys are gene IDs and values are
            gene annotation models.
        transcripts (Dict[str, TranscriptAnnotationModel]): Keys are transcript
            IDs and values are transcript annotation models.
        source (str): Source of the genomic annotation. E.g., ENSEMBL or
            GENCODE.
    """
    def __init__(self, genes:Dict[str,GeneAnnotationModel]=None,
            transcripts:Dict[str, TranscriptAnnotationModel]=None,
            source:str=None):
        """ Construct a GenomicAnnotation object """
        self.genes = genes or {}
        self.transcripts = transcripts or {}
        self.source = source
        self.gene_id_version_mapper:Dict[str, str] = None
        self.version = MetaVersion()
        self._cached_tx_seqs = []

    def __repr__(self) -> str:
        """ Return a string representation """
        result = ""
        i = 0
        it = iter(self.genes.keys())
        gene_id = next(it, None)
        while i < len(self.genes) and gene_id:
            result += f"'{gene_id}': {self.genes[gene_id].__repr__()}"
            if i == 3 and len(self.genes) > 7:
                result += "...\n"
                i = len(self.transcripts) - 4
            i += 1
            gene_id = next(it, None)
        result += f'\n{len(self.genes)} genes'

        i = 0
        it = iter(self.transcripts.keys())
        key = next(it, None)
        while i < len(self.transcripts) and key:
            result += f"'{key}': {self.transcripts[key].__repr__()}\n"
            if i == 3 and len(self.transcripts) > 7:
                result += "...\n"
                i = len(self.transcripts) - 4
            i += 1
            key = next(it, None)
        result += f'\n{len(self.transcripts)} transcripts'
        return result

    def add_gene_record(self, record:GTFSeqFeature) -> None:
        """ Add a gene record """
        gene_id = record.gene_id
        if gene_id in self.genes:
            raise ValueError(f'Same gene has multiple records: {gene_id}')
        record.__class__ = GeneAnnotationModel
        record.exons = []
        record.transcripts = []
        self.genes[gene_id] = record

    def add_transcript_record(self, record:GTFSeqFeature) -> None:
        """ Add a transcript record """
        feature = record.type.lower()
        if feature not in GTF_FEATURE_TYPES:
            return
        transcript_id = record.transcript_id
        record.id = transcript_id
        if transcript_id not in self.transcripts.keys():
            self.transcripts[transcript_id] = TranscriptAnnotationModel()
        self.transcripts[transcript_id].add_record(feature, record)

        gene_id = record.gene_id
        if gene_id not in self.genes:
            raise ValueError(f'Gene ID {gene_id} not found')
        if transcript_id not in self.genes[gene_id].transcripts:
            self.genes[gene_id].transcripts.append(transcript_id)

    def add_cached_tx_seq(self, tx_id:str):
        """ Add tx_id of cached transcript sequences """
        self._cached_tx_seqs.append(tx_id)

    def remove_cached_tx_seq(self):
        """ Remove all cached tx sequences """
        for tx_id in self._cached_tx_seqs:
            self.transcripts[tx_id].remove_cached_seq()

    def dump_gtf(self, handle:Union[str, IO], biotype:List[str]=None, source:str=None)->None:
        """ Dump a GTF file into a GenomicAnnotation

        Args:
            path (str): Path to a GTF file.
            biotype (List[str]): The annotation biotype to keep. Features
                in a GTF can be annotated as protein_coding, miRNA, lncRNA,
                miRNA, etc.
        """
        record:GTFSeqFeature
        if not source:
            inferrer = GTFSourceInferrer()

        for record in GtfIO.parse(handle):
            if biotype is not None and record.biotype not in biotype:
                continue

            if not source:
                record.source = inferrer.infer(record)
            else:
                record.source = source

            feature = record.type.lower()
            if feature == 'gene':
                self.add_gene_record(record)
                continue

            self.add_transcript_record(record)


        if not source:
            source = inferrer.source
        else:
            self.source = source

        for transcript_model in self.transcripts.values():
            transcript_model.sort_records()

    def check_protein_coding(self, proteome:AminoAcidSeqDict,
            invalid_protein_as_noncoding:bool) -> None:
        """ Checks if each transcript is protein coding """
        for tx_id, tx_model in self.transcripts.items():
            if invalid_protein_as_noncoding:
                if tx_id in proteome:
                    if '*' in proteome[tx_id].seq:
                        proteome.pop(tx_id)
                        tx_model.is_protein_coding = False
                    else:
                        tx_model.is_protein_coding = True
                else:
                    tx_model.is_protein_coding = False
            else:
                tx_model.is_protein_coding = tx_id in proteome

    def coordinate_gene_to_transcript(self, index:int, gene:str,
            transcript:str) -> int:
        """ For a given gene position, find it the corresponding transcript
        position.

        Args:
            index (int): The gene position index.
            gene (str): The gene ID.
            transcript (str): The transcript ID to map to.
        """
        genomic_index = self.coordinate_gene_to_genomic(index, gene)
        gene_model = self.genes[gene]
        if transcript not in gene_model.transcripts:
            raise ValueError(f'The transcript {transcript} is not associated'
                f'with the gene {gene}')
        transcript_model = self.transcripts[transcript]

        return transcript_model.get_transcript_index(genomic_index)

    def coordinate_genomic_to_gene(self, index:int, gene:str) -> int:
        """ Get the gene coordinate from genomic coordinate. """
        gene_location = self.genes[gene].location
        if gene_location.strand == 1:
            return index - gene_location.start
        if gene_location.strand == -1:
            return gene_location.end - 1 - index
        raise ValueError("Don't know how to handle unstranded gene.")

    def coordinate_gene_to_genomic(self, index:int, gene:str) -> int:
        """ Get the genomic coordinate from gene coordinate """
        location = self.genes[gene].location
        if location.strand == 1:
            return location.start + index
        if location.strand == -1:
            return location.end - 1 - index
        raise ValueError("Don't know how to handle unstranded gene.")

    def coordinate_transcript_to_genomic(self, index:int, transcript:str) -> int:
        """ Get the genomic coordinate from transcript coordinate """
        tx_model = self.transcripts[transcript]
        tx_size = sum(len(x.location) for x in tx_model.exon)
        if tx_size < index:
            raise ValueError('Index out of range.')

        if tx_model.transcript.strand == 1:
            for exon in tx_model.exon:
                size = exon.location.end - exon.location.start
                if index < size:
                    return index + exon.location.start
                index -= size
        if tx_model.transcript.strand == -1:
            for exon in reversed(tx_model.exon):
                size = exon.location.end - exon.location.start
                if index < size:
                    return exon.location.end - 1 - index
                index -= size
        raise ValueError("Don't know how to handle unstranded transcript.")

    def variant_coordinates_to_gene(self, variant:seqvar.VariantRecord,
            gene_id:str) -> seqvar.VariantRecord:
        """ Convert the coordinates of variant from transcript to gene.

        Note: if `variant.is_end_inclusion()` is true, which means the variant
        is a deletion and the first nucleotide of the deleted sequence is the
        first nucleotide after start codon, the variant is converted that the
        reference nucleotide becomes the next nucleotide of the deleted sequence
        from the gene sequence. This is to handle the case that the last
        nucleotide happens to be the end of an exon.
        """
        transcript_id = variant.location.seqname
        start = variant.location.start
        end = variant.location.end

        if transcript_id not in self.genes[gene_id].transcripts:
            raise ValueError("The variant isn't associated with the gene.")

        transcript_model = self.transcripts[transcript_id]
        start_genomic = self.coordinate_transcript_to_genomic(start, transcript_id)
        if variant.is_end_inclusion() and variant.type == 'INDEL' and variant.is_deletion():
            end_genomic = self.coordinate_transcript_to_genomic(end - 2, transcript_id)
            if transcript_model.transcript.strand == 1:
                end_genomic += 1
            else:
                end_genomic -= 1
        else:
            end_genomic = self.coordinate_transcript_to_genomic(end - 1, transcript_id)
        if transcript_model.transcript.strand == -1:
            start_genomic, end_genomic = end_genomic, start_genomic
        end_genomic += 1
        start_gene = self.coordinate_genomic_to_gene(start_genomic, gene_id)
        end_gene = self.coordinate_genomic_to_gene(end_genomic - 1, gene_id)
        if transcript_model.transcript.strand == -1:
            start_gene, end_gene = end_gene, start_gene
        end_gene += 1

        ref = variant.ref
        if end_gene - start_gene != end - start:
            if not (variant.type == 'INDEL' and variant.is_deletion()):
                raise ValueError(
                    'The variant is spanning over at least one entire intron with'
                    f"the type of '{variant.type}'. Don't know how to handle it."
                )
            ref = variant.id.split('-')[2]

        attrs = copy.copy(variant.attrs)
        if 'GENE_ID' in attrs:
            attrs.pop('GENE_ID')
        attrs['TRANSCRIPT_ID'] = transcript_id

        return seqvar.VariantRecord(
            location=FeatureLocation(
                seqname=gene_id,
                start=start_gene,
                end=end_gene
            ),
            ref=ref,
            alt=variant.alt,
            _type=variant.type,
            _id=variant.id,
            attrs=attrs
        )

    def feature_coordinate_gene_to_genomic(self, feature:SeqFeature,
            gene_id:str) -> SeqFeature:
        """ For a given feature, converts the coordinate from gene to
        genomic. """
        gene_model = self.genes[gene_id]
        start = self.coordinate_gene_to_genomic(feature.location.start, gene_id)
        end = self.coordinate_gene_to_genomic(feature.location.end - 1, gene_id)
        strand = gene_model.location.strand
        if strand == -1:
            start, end = end, start
        end += 1
        location = FeatureLocation(seqname=gene_id, start=start, end=end,
            strand=strand)
        new_feature = feature.__class__(
            chrom=gene_id, location=location, attributes={}
        )
        if hasattr(feature, 'source'):
            new_feature.source = feature.source
        return new_feature

    def feature_coordinate_genomic_to_gene(self, feature:SeqFeature,
            gene_id:str) -> SeqFeature:
        """ For a given feature, converts the coordinate from genomic to gene.
        """
        gene_model = self.genes[gene_id]
        start = self.coordinate_genomic_to_gene(feature.location.start, gene_id)
        end = self.coordinate_genomic_to_gene(feature.location.end - 1, gene_id)
        strand = gene_model.location.strand
        if strand == -1:
            start, end = end, start
        end += 1
        location = FeatureLocation(seqname=gene_id, start=start, end=end,
            strand=0)

        new_feature =  feature.__class__(
            chrom=gene_id, location=location, attributes={}
        )
        if hasattr(feature, 'source'):
            new_feature.source = feature.source

        return new_feature

    def create_gene_id_version_mapper(self) -> None:
        """ Create the a dict that keys are the unversioned gene ID, and values
        are the versioned. """
        self.gene_id_version_mapper = {}
        for versioned in self.genes.keys():
            unversioned = versioned.split('.')[0]
            if unversioned in self.gene_id_version_mapper:
                if '_PAR_Y' in versioned:
                    continue
                if '_PAR_Y' not in self.gene_id_version_mapper[unversioned]:
                    raise ValueError('Unversioned gene ID collapsed.')
            self.gene_id_version_mapper[unversioned] = versioned

    def get_gene_model_from_unversioned_id(self, gene_id:str) -> GeneAnnotationModel:
        """ Get the gene annotation model from an unversioned gene ID. In
        general the source and version of the annotation GTF file used should
        be consistent with the one used to call genomic variants (e.g. VEP &
        fusion). But this is useful for FusionCatcher because it uses ENSEMBL
        as default.

        Args:
            gene_id (str): The unversioned gene ID (ENSEMBL's style).

        Returns:
            The GeneAnnotationModel object of the gene.
        """
        if self.source == 'ENSEMBL':
            try:
                return self.genes[gene_id]
            except KeyError as error:
                raise err.GeneNotFoundError(gene_id) from error

        if self.gene_id_version_mapper is None:
            self.create_gene_id_version_mapper()

        try:
            versioned_gene_id = self.gene_id_version_mapper[gene_id]
        except KeyError as error:
            raise err.GeneNotFoundError(gene_id) from error

        return self.genes[versioned_gene_id]

    def get_all_exons_of_gene(self, gene_id) -> List[GTFSeqFeature]:
        """ Get all exons of a gene """
        exons = set()
        gene_model = self.genes[gene_id]
        for tx_id in gene_model.transcripts:
            tx_model = self.transcripts[tx_id]
            exons.update(tx_model.exon)
        exons = list(exons)
        exons.sort()
        gene_model.exons = exons

    def find_exon_index(self, transcript_id:str, feature:GTFSeqFeature,
            coordinate:str='gene') -> int:
        """ Find the exon index of the gene.

        Args:
            gene_id (str): The gene ID.
            feature (GTFSeqFeature): The exon to look up.
            coordinate (str): The coordinate type of the given feature. Can be
                one of 'genomic' or 'gene'. Defaults to gene.
        """
        tx_model = self.transcripts[transcript_id]
        gene_id = tx_model.transcript.gene_id
        exons = tx_model.exon

        strand = tx_model.transcript.strand

        if coordinate == 'gene':
            feature = self.feature_coordinate_gene_to_genomic(feature, gene_id)
        elif coordinate != 'genomic':
            raise ValueError("Don't know how to handle the coordinate.")

        if strand == 1:
            for i, exon in enumerate(exons):
                if exon == feature:
                    return i
                if exon > feature:
                    break
        elif strand == -1:
            for i, exon in enumerate(reversed(exons)):
                if exon == feature:
                    return i
                if exon < feature:
                    break
        raise err.ExonNotFoundError(gene_id, feature)

    def find_intron_index(self, transcript_id:str, feature:GTFSeqFeature,
            coordinate:str="gene", intron_start_range:Tuple[int,int]=(0,0),
            intron_end_range:Tuple[int,int]=(0,0)) -> int:
        """ Find the intron index of the gene.

        Args:
            gene_id (str): The gene ID.
            feature (GTFSeqFeature): The intron to look up.
            coordinate (str): The coordinate type of the given feature. Can be
                one of 'genomic' or 'gene'. Defaults to gene.
        """
        tx_model = self.transcripts[transcript_id]
        exons = tx_model.exon
        strand = tx_model.transcript.strand
        gene_id = tx_model.transcript.gene_id

        intron_start_range = FeatureLocation(
            start=intron_start_range[0],
            end=intron_start_range[1] + 1
        )
        intron_end_range = FeatureLocation(
            start=intron_end_range[0],
            end=intron_end_range[1] + 1
        )

        if coordinate == 'gene':
            feature = self.feature_coordinate_gene_to_genomic(feature, gene_id)

        elif coordinate != 'genomic':
            raise ValueError("Don't know how to handle the coordinate.")

        if strand == 1:
            it = enumerate(exons)
            while True:
                i, exon = next(it, (None, None))
                if not exon:
                    break
                start_offset = feature.location.start - exon.location.end
                if start_offset in intron_start_range:
                    i, exon = next(it, (None, None))
                    if not exon:
                        break
                    end_offset = feature.location.end - exon.location.start
                    if end_offset in intron_end_range:
                        return i - 1
                    if exon.location.start >= feature.location.end:
                        return i - 1
                    break
                if exon.location.start > feature.location.end:
                    break
        elif strand == -1:
            it = reversed(list(enumerate(exons)))
            while True:
                i, exon = next(it, (None, None))
                if not exon:
                    break
                start_offset = - (feature.location.end - exon.location.start)
                if start_offset in intron_start_range:
                    i, exon = next(it, (None, None))
                    if not exon:
                        break
                    end_offset = - (feature.location.start - exon.location.end)
                    if end_offset in intron_end_range:
                        return i + 1
                    if exon.location.end <= feature.location.start:
                        return i + 1
                    break
                if exon.location.end < feature.location.start:
                    break
        raise err.IntronNotFoundError(gene_id, feature)

    def get_transcripts_with_exonic_position(self, gene_id:str, pos:int
            ) -> List[TranscriptAnnotationModel]:
        """ get all transcripts of a gene that the given genomic position is
        exonic """
        transcripts:List[TranscriptAnnotationModel] = []
        for tx_id in self.genes[gene_id].transcripts:
            tx_model = self.transcripts[tx_id]
            if tx_model.is_exonic(pos):
                transcripts.append(tx_model)
        return transcripts

    def get_transcripts_with_position(self, gene_id:str, pos:str
            ) -> List[TranscriptAnnotationModel]:
        """ Get all transcripts of a gene that contains the genomic posision in
        exon or intron """
        transcripts:List[TranscriptAnnotationModel] = []
        for tx_id in self.genes[gene_id].transcripts:
            tx_model = self.transcripts[tx_id]
            if not tx_model.exon:
                continue
            start = tx_model.transcript.location.start
            end = tx_model.transcript.location.end
            if start <= pos < end:
                transcripts.append(tx_model)
        return transcripts

    def get_genes_rank(self) -> Dict[str, int]:
        """ Get the genes rank """
        i = 0
        rank = {}
        for gene_id in self.genes:
            rank[gene_id] = i
            i += 1
        return rank

    def get_transcript_rank(self) -> Dict[str, int]:
        """ Get the transcripts rank """
        i = 0
        rank = {}
        for tx_id in self.transcripts:
            rank[tx_id] = i
            i += 1
        return rank



""" `filterFasta` takes the FASTA file of variant peptides (output by
`callVariant`) or novel ORF peptides (output by `callNovelORF`) and filters it
based on the gene expression data. A expresion table must be given as a CSV
or TSV. """
from __future__ import annotations
import argparse
from pathlib import Path
import pickle
from typing import IO, Dict, List
from Bio import SeqIO
from moPepGen import get_logger
from moPepGen.aa import VariantPeptidePool
from moPepGen.aa.expasy_rules import EXPASY_RULES
from moPepGen.gtf.GenomicAnnotation import GenomicAnnotation
from moPepGen.cli import common


INPUT_FILE_FORMATS = ['.fasta', '.fa']
OUTPUT_FILE_FORMATS = ['.fasta', '.fa']

# pylint: disable=W0212
def add_subparser_filter_fasta(subparser:argparse._SubParsersAction):
    """ CLI for moPepGen filterFasta """
    p:argparse.ArgumentParser = subparser.add_parser(
        name='filterFasta',
        help='Filter noncanonical peptides.',
        description='Filter noncanonical peptides according to gene expression'
        ' or gene biotypes.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    common.add_args_input_path(
        parser=p, formats=INPUT_FILE_FORMATS,
        message='Input FASTA file, must be generated by either moPepGen'
        ' callVariant or callNovelORF.'
    )
    common.add_args_output_path(p, OUTPUT_FILE_FORMATS)
    p.add_argument(
        '--denylist',
        help='Path to the peptide sequence deny list. When using novel ORF'
        ' peptides as denylist, make sure it is no also passed as a input FASTA'
        ' file, because all peptide sequences will be removed. Valid formats:'
        ' [".fasta"]',
        type=Path,
        metavar='<file>',
        default=None
    )
    p.add_argument(
        '--exprs-table',
        type=Path,
        help="Path to the RNAseq quantification results.",
        metavar='<file>'
    )
    p.add_argument(
        '--skip-lines',
        type=int,
        help='Number of lines to skip when reading the expression table.'
        'Defaults to 0',
        metavar='<value>',
        default=0
    )
    p.add_argument(
        '--delimiter',
        type=str,
        help='Delimiter of the expression table. Defaults to tab.',
        metavar='<value>',
        default='\t'
    )
    p.add_argument(
        '--tx-id-col',
        type=str,
        help='The index for transcript ID in the RNAseq quantification results.'
        ' Index is 1-based.',
        metavar='<number>'
    )
    p.add_argument(
        '--quant-col',
        type=str,
        help='The column index number for quantification. Index is 1-based.',
        metavar='<number>'
    )
    p.add_argument(
        '--quant-cutoff',
        type=float,
        help='Quantification cutoff.',
        metavar='<number>',
        default=None
    )
    p.add_argument(
        '--keep-all-coding',
        action='store_true',
        default=False,
        help='Keep all coding genes, regardless of their expression level.'
    )
    p.add_argument(
        '--keep-all-noncoding',
        action='store_true',
        default=False,
        help='Keep all noncoding genes, regardless of their expression level.'
    )
    p.add_argument(
        '--keep-canonical',
        action='store_true',
        help='Keep peptides called from canonical ORFs. Only useful together'
        ' with denylist.'
    )
    p.add_argument(
        '--miscleavages',
        type=str,
        help='Range of miscleavages per peptide to allow. Min and max are'
        ' included. For example, "1:2" will keep all peptides with 1 or 2'
        ' miscleavages.',
        metavar='[min]:[max]',
        default=None
    )
    p.add_argument(
        '--enzyme',
        type=str,
        help='Enzyme name. Ignored if --miscleavages is not specified.',
        choices=list(EXPASY_RULES.keys()),
        default='trypsin'
    )

    common.add_args_reference(p, genome=False, proteome=False)
    common.add_args_debug_level(p)
    common.print_help_if_missing_args(p)
    p.set_defaults(func=filter_fasta)
    return p

def filter_fasta(args:argparse.Namespace) -> None:
    """ Filter noncanonical peptide FASTA """
    logger = get_logger()
    common.validate_file_format(
        args.input_path, INPUT_FILE_FORMATS, check_readable=True
    )
    common.validate_file_format(
        args.output_path, OUTPUT_FILE_FORMATS, check_writable=True
    )

    if args.miscleavages is not None and not ':' in args.miscleavages:
        raise ValueError('Value invalid for --miscleavages. Must be [min]:[max]')

    if args.miscleavages is None:
        miscleavage_range = (None, None)
    else:
        miscleavage_range = tuple(int(x) for x in args.miscleavages.split(':', 1))

    common.print_start_message(args)

    coding_tx = load_coding_transcripts(args)

    with open(args.input_path, 'rt') as handle:
        pool = VariantPeptidePool.load(handle)

    logger.info('Peptide FASTA file loaded.')

    if args.exprs_table is not None:
        with open(args.exprs_table, 'rt') as handle:
            i = 0
            while i < args.skip_lines:
                handle.readline()
                i += 1
            tx_id_col:str = args.tx_id_col
            quant_col:str = args.quant_col
            if any(not x.isdecimal() for x in [tx_id_col, quant_col]):
                header = handle.readline().rstrip().split(args.delimiter)

            if tx_id_col.isdecimal():
                tx_id_col = int(tx_id_col) - 1
            else:
                tx_id_col = header.index(tx_id_col)

            if quant_col.isdecimal():
                quant_col = int(quant_col) - 1
            else:
                quant_col = header.index(quant_col)

            exprs = load_expression_table(
                handle=handle,
                tx_col=tx_id_col,
                quant_col=quant_col,
                delim=args.delimiter
            )
        logger.info('Gene expression table loaded.')
    else:
        exprs = None

    if args.denylist is not None:
        with open(args.denylist, 'rt') as handle:
            denylist = {seq.seq for seq in SeqIO.parse(handle, 'fasta')}
        logger.info('Peptide denylist loaded.')
    else:
        denylist = None

    filtered_pool = pool.filter(
        exprs=exprs, cutoff=args.quant_cutoff, coding_transcripts=coding_tx,
        keep_all_noncoding=args.keep_all_noncoding,
        keep_all_coding=args.keep_all_coding, enzyme=args.enzyme,
        miscleavage_range=miscleavage_range, denylist=denylist,
        keep_canonical=args.keep_canonical
    )

    filtered_pool.write(args.output_path)
    logger.info('Filtered FASTA file saved.')

def load_expression_table(handle:IO, tx_col:int,quant_col:int,
        delim:str='\t') -> Dict[str,float]:
    """ Load the gene expression quantification table """
    data = {}
    line:str
    for line in handle:
        fields = line.rstrip().split(delim)
        tx_id = fields[tx_col]
        quant = float(fields[quant_col])
        data[tx_id] = quant

    return data

def load_coding_transcripts(args:argparse.Namespace) -> List[str]:
    """ load and get the protein coding transcripts """
    if args.index_dir:
        with open(args.index_dir/'coding_transcripts.pkl', 'rb') as handle:
            return pickle.load(handle)

    anno = GenomicAnnotation()
    anno.dump_gtf(args.annotation_gtf)

    return {tx_id for tx_id, tx_model in anno.transcripts.items()
        if tx_model.is_protein_coding}



""" This module takes in a variant peptide FASTA file and encodes the fasta
headers with 36-digit long UUID (32 digits of hexdecimal characters and 4
hypens). The original headers together with  the UUIDs arer saved into the
dict file at the same location. This resolves the problems that some proteomic
search engines have strick requirement on the FASTA header length. """
from __future__ import annotations
import argparse
from pathlib import Path
from typing import TYPE_CHECKING, Dict
import uuid
from Bio import SeqIO
from Bio.SeqIO import FastaIO
from moPepGen.cli import common


if TYPE_CHECKING:
    from Bio.SeqRecord import SeqRecord

INPUT_FILE_FORMATS = ['.fa', '.fasta']
OUTPUT_FILE_FORMATS = ['.fa', '.fasta']

# pylint: disable=W0212
def add_subparser_encode_fasta(subparser:argparse._SubParsersAction):
    """ CLI for moPepGen encodeFasta """
    parser:argparse.ArgumentParser = subparser.add_parser(
        name='encodeFasta',
        help='Encode variant peptide FASTA file header.',
        description='Encode variant peptide FASTA file header.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    common.add_args_input_path(
        parser=parser, formats=INPUT_FILE_FORMATS,
        message='Input FASTA file, must be generated by moPepGen callVariant.'
    )
    common.add_args_output_path(
        parser=parser, formats=OUTPUT_FILE_FORMATS
    )
    common.add_args_decoy(parser)
    common.add_args_debug_level(parser)
    common.print_help_if_missing_args(parser)
    parser.set_defaults(func=encode_fasta)
    return parser

def encode_fasta(args:argparse.Namespace) -> None:
    """ encode fasta """
    common.print_start_message(args)

    input_path:Path = args.input_path
    output_path:Path = args.output_path
    common.validate_file_format(
        input_path, INPUT_FILE_FORMATS, check_readable=True
    )
    common.validate_file_format(
        output_path, OUTPUT_FILE_FORMATS, check_writable=False
    )

    fasta_dict = output_path.with_suffix(output_path.suffix + '.dict')

    with open(input_path, 'rt') as in_handle, \
            open(output_path, 'wt') as out_handle, \
            open(fasta_dict, 'wt') as dict_handle:
        record2title = lambda x: x.description
        id_mapper:Dict[str,str] = {}
        writer = FastaIO.FastaWriter(out_handle, record2title=record2title)
        record:SeqRecord
        for record in SeqIO.parse(in_handle, 'fasta'):
            is_decoy = is_decoy_sequence(record.description, args.decoy_string,
                args.decoy_string_position)
            if is_decoy:
                header = get_real_header(record.description, args.decoy_string,
                    args.decoy_string_position)
            else:
                header = record.description

            if header in id_mapper:
                index = id_mapper[header]
            else:
                index = str(uuid.uuid4())
                dict_handle.write(f"{index}\t{header}" + '\n')
                id_mapper[header] = index

            if is_decoy:
                index = get_decoy_header(index, args.decoy_string,
                    args.decoy_string_position)

            record.description = index
            writer.write_record(record)

def is_decoy_sequence(header:str, decoy_string:str, decoy_string_position:str) -> bool:
    """ Checks if the sequence is decoy """
    if decoy_string_position == 'prefix':
        return header.startswith(decoy_string)
    if decoy_string_position == 'suffix':
        return header.endswith(decoy_string)
    raise ValueError(f"decoy_string_position of {decoy_string_position} is not supported.")

def get_real_header(header:str, decoy_string:str, decoy_string_position:str) -> str:
    """ Get the real header if it is a decoy sequence """
    if decoy_string_position == 'prefix':
        return header[len(decoy_string):]
    if decoy_string_position == 'suffix':
        return header[:-len(decoy_string)]
    raise ValueError(f"decoy_string_position of {decoy_string_position} is not supported.")

def get_decoy_header(header:str, decoy_string:str, decoy_string_position:str) -> str:
    """ Get the decoyed version of the header """
    if decoy_string_position == 'prefix':
        return decoy_string + header
    if decoy_string_position == 'suffix':
        return header + decoy_string
    raise ValueError(f"decoy_string_position of {decoy_string_position} is not supported.")



""" SubgraphMapper """
from __future__ import annotations
from typing import Dict, Tuple, TYPE_CHECKING
import uuid
from moPepGen.SeqFeature import FeatureLocation


if TYPE_CHECKING:
    from moPepGen.seqvar import VariantRecord

class SubgraphBranch():
    """ ParentGraphLocation """
    def __init__(self, _id:str, level:int, parent_id:str,  location:FeatureLocation,
            variant:VariantRecord, feature_type:str, feature_id:str):
        """ constructur """
        self.id = _id
        self.level = level
        self.parent_id = parent_id
        self.location = location
        self.variant = variant
        self.feature_type = feature_type
        self.feature_id= feature_id

    def is_parent(self, other:SubgraphBranch) -> bool:
        """ Checks if it is the parent of a given subgraph """
        return self.id == other.parent_id

class SubgraphTree():
    """ This SubgraphTree class defines the relationship of subgraphes and
    parents/children in a tree-like data structure. """
    def __init__(self, data:Dict[str, SubgraphBranch]=None,
            root:SubgraphBranch=None):
        """ Constructor """
        self.data = data or {}
        self.root = root

    def __getitem__(self, key:str) -> SubgraphBranch:
        """ Get item """
        return self.data[key]

    def add_root(self, _id:str, feature_type:str, feature_id:str, variant:VariantRecord):
        """ Add the main graph """
        subgraph = SubgraphBranch(
            _id=_id, level=0, parent_id=None, location=None, variant=variant,
            feature_type=feature_type, feature_id=feature_id
        )
        self.data[_id] = subgraph
        self.root = subgraph

    def add_subgraph(self, child_id:str, parent_id:str, level:int,
            start:int, end:int, variant:VariantRecord,
            feature_type:str, feature_id:str):
        """ Add subgraph """
        location = FeatureLocation(
            seqname=parent_id, start=start, end=end
        )
        subgraph = SubgraphBranch(
            _id=child_id, level=level, parent_id=parent_id,
            location=location, variant=variant,
            feature_type=feature_type, feature_id=feature_id
        )
        self.data[child_id] = subgraph

    @staticmethod
    def generate_subgraph_id() -> str:
        """ Generate a random subgraph ID """
        return str(uuid.uuid4())

    def find_compatible_parents(self, first:str, second:str
            ) -> Tuple[SubgraphBranch, SubgraphBranch]:
        """ For two given subgraphs, find their parent graphs that are
        compatible. Compatible means either they share the same parent subgraph,
        or one is the parent of the other subgraph. """
        subgraph1 = self[first.subgraph_id]
        subgraph2 = self[second.subgraph_id]

        while subgraph1.id != subgraph2.id and \
                subgraph1.parent_id != subgraph2.parent_id and \
                subgraph1.id !=  subgraph2.parent_id and \
                subgraph1.parent_id != subgraph2.id:
            level1 = subgraph1.level
            level2 = subgraph2.level
            if level1 >= level2:
                subgraph1 = self[subgraph1.parent_id]
            if level2 >= level1:
                subgraph2 = self[subgraph2.parent_id]

        return subgraph1, subgraph2



""" Module for GVF metadata """
from __future__ import annotations
from typing import List, IO
from moPepGen import __version__, constant
from .GVFMetadataInfo import GVF_METADATA_INFO, GVF_METADATA_ADDITIONAL


ALT_DESCRIPTION = {
    'DEL': 'Deletion relative to the reference',
    'INS': 'Insertion of novel sequence relative to the reference',
    'DUP': 'Region of elevated copy number relative to the reference',
    'INV': 'Inversion of reference sequence',
    'DUP:TANDEM': 'Tandem duplication',
    'DEL:ME': 'Deletion of mobile element relative to the reference',
    'INS:ME': 'Insertion of a mobile element relative to the reference',
    'FUSION': 'Gene fusion'
}

class GVFMetadata():
    """ A GVFMetadata object contains the metadata section to be written into
    a GVF file.

    Attributes:
        parser (str): The moPepGen parser
        reference_index (str): Path to the reference index directory.
        genome_fasta (str): Path to the genome fasta file.
        annotation_gtf (str): Path to the annotation GTF file.
        alt (dict): Alternative allele fields.
        info (dict): Information fields.
    """
    def __init__(self, parser:str, source:str, chrom:str,
            reference_index:str=None, genome_fasta:str=None,
            annotation_gtf:str=None, version:str=None, info=None,
            additional=None):
        """ Construct a TVFMetadata object. """
        self.parser = parser
        self.source = source
        self.chrom = chrom
        self.reference_index = reference_index
        self.genome_fasta = genome_fasta
        self.annotation_gtf = annotation_gtf
        self.alt = {}
        self.info = info or GVF_METADATA_INFO['Base']
        self.added_types = []
        self.additional = additional
        self.version = version or __version__

    def add_info(self, variant_type:str) -> None:
        """ Add a INFO field to the metadata. """
        self.add_alt(variant_type)
        if variant_type in self.added_types:
            return
        if variant_type in constant.SINGLE_NUCLEOTIDE_SUBSTITUTION:
            return
        if variant_type == 'Fusion':
            self.info.update(GVF_METADATA_INFO['Fusion'])
            self.added_types.append('Fusion')
        elif variant_type == 'Insertion':
            self.info.update(GVF_METADATA_INFO['Insertion'])
            self.added_types.append('Insertion')
        elif variant_type == 'Deletion':
            self.info.update(GVF_METADATA_INFO['Deletion'])
            self.added_types.append('Deletion')
        elif variant_type == 'Substitution':
            self.info.update(GVF_METADATA_INFO['Substitution'])
            self.added_types.append('Substitution')
        elif variant_type == 'circRNA':
            self.info.update(GVF_METADATA_INFO['circRNA'])
            self.added_types.append('circRNA')
            self.additional = GVF_METADATA_ADDITIONAL['circRNA']
        else:
            raise ValueError(f'Unknown variant type: {variant_type}')

    def add_alt(self, variant_type:str) -> None:
        """ Add a ALT field to the metadata. """
        if variant_type in ALT_DESCRIPTION and variant_type not in self.alt:
            self.alt[variant_type] = ALT_DESCRIPTION[variant_type]

    def to_strings(self) -> List[str]:
        """ Convert metadata to a list of strings to be written to a GVF file. """
        info_lines = []
        for key,val in self.info.items():
            number = val['Number']
            _type = val['Type']
            desc = val['Description']
            line = f'##INFO=<ID={key},Number={number},Type={_type},Description="{desc}">'
            info_lines.append(line)

        additional = []
        if self.additional:
            for key,val in self.additional.items():
                values = []
                for k,v in val.items():
                    values.append(f"{k}={v}")
                additional.append(f"##{key}=<{','.join(values)}>")

        ref_index = self.reference_index if self.reference_index else ''
        genome_fasta = self.genome_fasta if self.genome_fasta else ''
        annotation_gtf = self.annotation_gtf if self.annotation_gtf else ''
        return [
            '##fileformat=VCFv4.2',
            f'##mopepgen_version={self.version}',
            f'##parser={self.parser}',
            f'##reference_index={ref_index}',
            f'##genome_fasta={genome_fasta}',
            f'##annotation_gtf={annotation_gtf}',
            f'##source={self.source}',
            f'##CHROM=<Description="{self.chrom}">',
            *[f'##ALT=<ID={key},Description="{val}">' for key, val in self.alt],
            *info_lines,
            *additional
        ]

    def is_circ_rna(self) -> bool:
        """ checks if this is a circRNA """
        return self.parser in ['parseCIRCexplorer']

    @classmethod
    def parse(cls, handle:IO) -> GVFMetadata:
        """ Parse the metadata from a GVF file """
        pos = handle.tell()
        it = handle.readline()
        metadata = {}
        while it and it.startswith('##'):
            key, val = it.lstrip('##').rstrip().split('=', 1)
            if val == '':
                val = None
            if key == 'CHROM':
                key = key.lower()
                for it in val.strip('<>').split(','):
                    k,v = it.split('=')
                    v = v.strip('"').strip("'")
                    if k == 'Description':
                        metadata[key] = v
            elif key == 'INFO':
                key = key.lower()
                if key not in metadata:
                    metadata[key] = {}
                info_key = None
                info_val = {}
                for it in val.strip('<>').split(','):
                    k,v = it.split('=')
                    v = v.strip('"')
                    if k == 'ID':
                        info_key = v
                    else:
                        info_val[k] = v
                if info_key is None:
                    raise ValueError('Could not find the key.')
                metadata[key][info_key] = info_val
            elif key in ['ALT', 'POS']:
                pass
            elif key == 'mopepgen_version':
                metadata['version'] = val
            elif key != 'fileformat':
                metadata[key] = val
            pos = handle.tell()
            it = handle.readline()
        handle.seek(pos)
        return cls(**metadata)



"""Main entry point for moPepGen-util """
import sys
import argparse
from moPepGen import __version__, util
from moPepGen.cli.common import setup_loggers


def main():
    """ Main entry point """

    parser = argparse.ArgumentParser(
        prog='moPopGen-util'
    )
    parser.add_argument(
        '-V', '--version',
        help='Show version number and exit.',
        action='store_true'
    )
    subparsers = parser.add_subparsers(
        dest='command'
    )

    util.brute_force.parse_args(subparsers)
    util.brute_force_novel_orf.parse_args(subparsers)
    util.downsample_reference.parse_args(subparsers)
    util.validate_variant_calling.parse_args(subparsers)
    util.fuzz_test.parse_args(subparsers)
    util.extract_gvf.parse_args(subparsers)
    util.validate_novel_orf_calling.parse_args(subparsers)
    util.add_fuzz_test_log.parse_args(subparsers)

    args = parser.parse_args()

    if len(sys.argv)==1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    setup_loggers(args.debug_level)

    if args.version:
        print(f'moPepGen {__version__}', file=sys.stdout, flush=True)
        sys.exit()

    args.func(args)


if __name__ == '__main__':
    main()



""" Module for node in the peptide graph """
from __future__ import annotations
import copy
from functools import cmp_to_key
from collections import deque
import math
from typing import Dict, List, Set, Tuple, Iterable
from moPepGen import aa, circ, seqvar, get_logger
from moPepGen.SeqFeature import FeatureLocation
from moPepGen.seqvar.VariantRecord import VariantRecord
from moPepGen.svgraph.SubgraphTree import SubgraphTree


class PVGNode():
    """ The PVGNode class is used in the svgraph.PeptideVariantGraph, that
    stores the sequence in a Node.

    Attributes:
        - `seq` (AminoAcidSeqRecord): The amino acid sequence.
        - `reading_frame_index` (int): Reading frame index that the node belongs
          to. Must be 0, 1, or 2.
        - `variants` (List[seqvar.VariantRecordWithCoordinate]): The variant
          records carried in the sequence.
        - `in_nodes` (Set[PVGNode]): Inbound nodes.
        - `out_nodes` (Set[PVGNode]): Outbound nodes
        - `cleavage` (bool): Whether the start of the node is a cleavage site.
        - `truncated` (bool): Whether the node is truncated. Useful when the
          sequence does not have a confirmed stop codon.
        - `orf` (List[int]): List of two integers indicating the start and stop
          of the orf.
        - `was_bridge` (bool): When true indicating that the node used to be a
          bridge node between reading frames, but is no longer a bridge node
          because it was split.
        - `pre_cleaved` (bool): Indicating that the node is already cleaved.
        - `level` (bool): Subgraph level that the node belongs to.
        - `npop_collapsed` (bool): Whether the n-terminus is pop collapsed.
        - `cpop_collapsed` (bool): Whether the c-terminus is pop collapsed.
        - `upstream_indel_map` (Dict[PVGNode, List[VariantRecord]]): A mapping
          from upstream node to indel variants. When a node contains INDELs is
          collapsed with a canonical peptide node, the upstream node will be
          called for miscleaved peptides that has corresponding INDELs.
    """
    def __init__(self, seq:aa.AminoAcidSeqRecordWithCoordinates,
            reading_frame_index:int, subgraph_id:str,
            variants:List[seqvar.VariantRecordWithCoordinate]=None,
            in_nodes:Set[PVGNode]=None, out_nodes:Set[PVGNode]=None,
            selenocysteines:List[seqvar.VariantRecordWithCoordinate]=None,
            cleavage:bool=False, truncated:bool=False, orf:List[int]=None,
            was_bridge:bool=False, pre_cleaved:bool=False, level:int=0,
            npop_collapsed:bool=False, cpop_collapsed:bool=False,
            upstream_indel_map:Dict[PVGNode, List[seqvar.VariantRecord]]=None,
            collapsed_variants:Dict[PVGNode, List[seqvar.VariantRecordWithCoordinate]]=None,
            left_cleavage_pattern_end:int=None, right_cleavage_pattern_start:int=None
            ):
        """ Construct a PVGNode object. """
        self.seq = seq
        self.variants = variants or []
        self.in_nodes = in_nodes or set()
        self.out_nodes = out_nodes or set()
        self.selenocysteines = selenocysteines or []
        self.cleavage = cleavage
        self.truncated = truncated
        self.orf = orf or [None, None]
        self.reading_frame_index = reading_frame_index
        self.was_bridge = was_bridge
        self.pre_cleave = pre_cleaved
        self.subgraph_id = subgraph_id
        self.level = level
        self.npop_collapsed = npop_collapsed
        self.cpop_collapsed = cpop_collapsed
        self.upstream_indel_map = upstream_indel_map or {}
        self.collapsed_variants = collapsed_variants or {}
        self.left_cleavage_pattern_end = left_cleavage_pattern_end
        self.right_cleavage_pattern_start = right_cleavage_pattern_start

    def __getitem__(self, index) -> PVGNode:
        """ get item """
        start, stop, _ = index.indices(len(self.seq))
        location = FeatureLocation(start=start, end=stop)
        seq = self.seq.__getitem__(index)
        variants = []

        for variant in self.variants:
            if variant.location.overlaps(location):
                variants.append(variant.shift(-start))

        npop_collapsed = self.npop_collapsed and start == 0
        cpop_collapsed = self.cpop_collapsed and stop == len(seq) or stop == -1

        if start == 0:
            upstream_indel_map = {k:copy.copy(v) for k,v in self.upstream_indel_map.items()}
        else:
            upstream_indel_map = []

        collapsed_variants = {}
        for upstream, variants in self.collapsed_variants.items():
            filtered_variants = []
            for v in variants:
                if v.location.overlaps(location):
                    v = v[start:stop]
                    v = v.shift(-start)
                    filtered_variants.append(v)
            if filtered_variants:
                collapsed_variants[upstream] = filtered_variants

        secs = []
        for sec in self.selenocysteines:
            if start <= sec.location.start < stop:
                secs.append(sec.shift(-start))

        if self.left_cleavage_pattern_end and start < self.left_cleavage_pattern_end:
            left_cleavage_pattern_end = self.left_cleavage_pattern_end - start
        else:
            left_cleavage_pattern_end = None

        if self.right_cleavage_pattern_start and stop > self.right_cleavage_pattern_start:
            right_cleavage_pattern_start = self.right_cleavage_pattern_start
        else:
            right_cleavage_pattern_start = None

        return PVGNode(
            seq=seq,
            variants=variants,
            cleavage=self.cleavage,
            orf=self.orf,
            selenocysteines=secs,
            reading_frame_index=self.reading_frame_index,
            was_bridge=self.was_bridge,
            subgraph_id=self.subgraph_id,
            level=self.level,
            npop_collapsed=npop_collapsed,
            cpop_collapsed=cpop_collapsed,
            upstream_indel_map=upstream_indel_map,
            collapsed_variants=collapsed_variants,
            left_cleavage_pattern_end=left_cleavage_pattern_end,
            right_cleavage_pattern_start=right_cleavage_pattern_start
        )

    def add_out_edge(self, node:PVGNode) -> None:
        """ Add a outbound edge from this node.

        Args:
            node (PVGNode): The outboud node of the edge to add.
        """
        self.out_nodes.add(node)
        node.in_nodes.add(self)

    def remove_out_edge(self, node:PVGNode) -> None:
        """ Remove a outbound edge. It tries to remove the node, but won't
        raise any error if it doesn't exist.

        Args:
            node (PVGNode): The outbound node of the edge to remove.
        """
        self.out_nodes.discard(node)
        node.in_nodes.discard(self)

    def get_out_nodes(self) -> List[PVGNode]:
        """ Get outgoing nodes as a list. """
        return list(self.out_nodes)

    def get_in_nodes(self) -> List[PVGNode]:
        """ Get incoming nodes as a list """
        return list(self.in_nodes)

    def has_exclusive_outbond_node(self) -> bool:
        """ The given node has exclusive outbond node """
        return len(self.out_nodes) == 1 and \
            len(self.get_out_nodes()[0].in_nodes) == 1

    def has_multiple_segments(self) -> bool:
        """ Whether the node has multiple segments, which is when the node
        is merged from several individual nodes. """
        n = len(self.seq.locations)
        for v in self.variants:
            if not (v.variant.is_fusion() \
                    or v.variant.is_circ_rna() \
                    or (v.variant.is_alternative_splicing() and not v.variant.is_deletion())):
                n += 1
        return n > 1

    def _get_nth_rf_index(self, i:int) -> int:
        """ """
        if (i > 0 or i < -1) and not self.has_multiple_segments():
            raise ValueError('Node does not have multiple segments')

        locations = [loc.query for loc in self.seq.locations]
        for v in self.variants:
            if not (v.variant.is_fusion() \
                    or v.variant.is_circ_rna() \
                    or (v.variant.is_alternative_splicing() and not v.variant.is_deletion()) \
                    or v.downstream_cleavage_altering \
                    or v.upstream_cleavage_altering):
                locations.append(v.location)

        locations.sort()

        return locations[i].reading_frame_index

    def get_first_rf_index(self) -> int:
        """ Get the first fragment's reading frame index """
        return self._get_nth_rf_index(0)

    def get_second_rf_index(self) -> int:
        """ Get the second fragment's reading frame index """
        return self._get_nth_rf_index(1)

    def get_last_rf_index(self) -> int:
        """ Get the last fragment's reading frame index """
        return self._get_nth_rf_index(-1)

    def _get_nth_subgraph_id(self, i) -> str:
        """ Get the nth fragment's subgraph ID """
        if (i > 0 or i < -1) and not self.has_multiple_segments():
            raise ValueError('Node does not have multiple segments')

        locations = [(loc.query, loc.ref.seqname) for loc in self.seq.locations]
        locations += [(v.location, v.location.seqname) for v in self.variants
            if not v.variant.is_circ_rna()]

        locations = sorted(locations, key=lambda x: x[0])

        return locations[i][1]

    def get_first_subgraph_id(self) -> str:
        """ Get the first fragment's subgraph ID """
        return self._get_nth_subgraph_id(0)

    def get_last_subgraph_id(self) -> str:
        """ Get the last fragment's subgraph ID """
        return self._get_nth_subgraph_id(-1)

    def get_first_dna_ref_position(self) -> int:
        """ Get the first DNA reference position """
        if self.seq.locations:
            first_seq_loc = self.seq.locations[0]
        else:
            first_seq_loc = None

        non_circ_variants = [v for v in self.variants if not v.variant.is_circ_rna()]
        if non_circ_variants:
            first_var_loc = non_circ_variants[0].location
        else:
            first_var_loc = None

        if not first_seq_loc and not first_var_loc:
            raise ValueError(f"The node (seq={str(self.seq.seq)}) does not have loc")

        if not first_var_loc or first_seq_loc.query < first_var_loc:
            return first_seq_loc.get_ref_dna_start()

        return non_circ_variants[0].variant.location.end

    def get_last_dna_ref_position(self) -> int:
        """ Get the first DNA reference position """
        if self.seq.locations:
            last_seq_loc = self.seq.locations[-1]
        else:
            last_seq_loc = None

        non_circ_variants = [v for v in self.variants if not v.variant.is_circ_rna()]
        if non_circ_variants:
            last_var_loc = non_circ_variants[-1].location
        else:
            last_var_loc = None

        if not last_seq_loc and not last_var_loc:
            raise ValueError(f"The node (seq={str(self.seq.seq)}) does not have loc")

        if not last_var_loc or last_seq_loc.query > last_var_loc:
            return last_seq_loc.get_ref_dna_end()

        return non_circ_variants[-1].variant.location.end

    def remove_out_edges(self) -> None:
        """ remove output nodes """
        for node in copy.copy(self.out_nodes):
            self.remove_out_edge(node)

    def is_inbond_of(self, node:PVGNode) -> bool:
        """ Checks if self is the inbond node of a given node. """
        return node in self.out_nodes

    def is_orphan(self) -> bool:
        """ Checks if the node is orphan (no inbond or outbond node) """
        return not self.in_nodes and not self.out_nodes

    def is_bridge(self) -> None:
        """ Check if this is a bridge node to another reading frame """
        for node in self.out_nodes:
            if not (not node.out_nodes and node.seq.seq == '*') and \
                    node.reading_frame_index != self.reading_frame_index:
                return True
        if self.was_bridge:
            return True
        return False

    def is_subgraph_end(self) -> bool:
        """ Check if this is the end node of a subgraph """
        return all(x.subgraph_id == self.subgraph_id for x in self.in_nodes) and \
            any(x.subgraph_id != self.subgraph_id for x in self.out_nodes)

    def is_subgraph_start(self) -> bool:
        """ Check if this is the start node of a subgraph """
        return any(x.subgraph_id != self.subgraph_id for x in self.in_nodes) and \
            all(x.subgraph_id == self.subgraph_id for x in self.out_nodes)

    def is_at_least_one_loop_downstream(self, other, subgraphs:SubgraphTree,
            circ_rna:circ.CircRNAModel) -> bool:
        """ Checks if is at least one loop downstream to the other node. """
        if self.seq.locations:
            i = self.seq.locations[-1].ref.start
            x_level = subgraphs[self.seq.locations[-1].ref.seqname].level
        else:
            for v in reversed(self.variants):
                if not v.variant.is_circ_rna():
                    i = math.floor(v.variant.location.start / 3)
                    x_level = subgraphs[v.location.seqname].level
                    break
            else:
                raise ValueError("Failed to find a non circRNA variant.")

        if other.seq.locations:
            j = other.seq.locations[0].ref.start
            y_level = subgraphs[other.seq.locations[0].ref.seqname].level
        else:
            for v in other.variants:
                if not v.variant.is_circ_rna():
                    j = math.floor(v.variant.location.start / 3)
                    y_level = subgraphs[v.location.seqname].level
                    break
            else:
                raise ValueError("Failed to find a non circRNA variant.")

        if x_level - y_level > 1:
            return True

        if x_level == y_level:
            return False

        if x_level - y_level == 1:
            for fragment in circ_rna.fragments:
                frag = FeatureLocation(
                    start=math.floor((fragment.location.start - 3) / 3),
                    end=math.floor(fragment.location.end / 3)
                )
                if i in frag and j in frag:
                    return i >= j
                if i in frag:
                    return False
                if j in frag:
                    return True
        raise ValueError('Locations not found from the fragments of the circRNA.')

    def has_branch_sequence_matches(self, p:str) -> bool:
        """ Checks whether the node has a series of outgoing nodes that has the
        given sequence. """
        queue = deque([('', self)])
        while queue:
            cur_seq, cur_node = queue.popleft()
            seq = cur_seq + str(cur_node.seq.seq)
            if len(seq) > len(p):
                continue
            if not p.startswith(seq):
                continue
            if len(seq) == len(p):
                return True
            for node in cur_node.get_out_nodes():
                queue.append((seq, node))
        return False

    def has_any_in_bridge(self) -> None:
        """ Check if it has any incoming node that is bridge """
        return any(node.is_bridge() for node in self.in_nodes)

    def has_any_indel(self) -> None:
        """ Checks if there is any indel """
        return any(x.variant.is_indel() for x in self.variants)

    def get_variants_at(self, start:int, end:int=-1,
            upstream_cleavage_altering:bool=True,
            downstream_cleavage_altering:bool=True
            ) -> List[seqvar.VariantRecord]:
        """ Get the variant at position i """
        if end == -1:
            end = len(self.seq)
        if not -1 < start <= end <= len(self.seq):
            raise ValueError('start and end out of the range')
        variants = []
        location = FeatureLocation(start=start, end=end)
        for variant in self.variants:
            if not upstream_cleavage_altering and variant.upstream_cleavage_altering:
                continue
            if not downstream_cleavage_altering and variant.downstream_cleavage_altering:
                continue
            if variant.location.overlaps(location):
                variants.append(variant.variant)
        return variants

    def get_cleavage_gain_variants(self) -> List[seqvar.VariantRecord]:
        """ Get cleavage gain variants """
        cleavage_gain = []
        for variant in self.variants:
            if self.right_cleavage_pattern_start is not None \
                    and variant.location.end >= self.right_cleavage_pattern_start \
                    and not variant.is_silent:
                cleavage_gain.append(variant.variant)
        return cleavage_gain

    def get_cleavage_gain_from_downstream(self) -> List[seqvar.VariantRecord]:
        """ Get the variants that gains the cleavage by downstream nodes
        Downstream variant altering variants are those that overlaps with the
        `left_cleavage_pattern_start` of the downstream node.
        """
        cleavage_gain = []

        if self.right_cleavage_pattern_start:
            upstream_cleave_alts = [v.variant for v in self.variants
                if v.location.end > self.right_cleavage_pattern_start]
        else:
            seq_len = len(self.seq.seq)
            upstream_cleave_alts = [v.variant for v in self.variants
                if v.location.end == seq_len]
        for node in self.out_nodes:
            if not node.left_cleavage_pattern_end:
                continue
            if not node.variants:
                return []
            if node.variants[0].location.start >= node.left_cleavage_pattern_end:
                return []
            for v in node.variants:
                if v.location.start >= node.left_cleavage_pattern_end:
                    break
                if not cleavage_gain and not v.is_silent \
                        and not v.downstream_cleavage_altering:
                    if v.upstream_cleavage_altering:
                        if not any(x == v.variant for x in upstream_cleave_alts):
                            continue
                    cleavage_gain.append(v.variant)
        return cleavage_gain

    def get_stop_lost_variants(self, stop_index:int) -> List[seqvar.VariantRecord]:
        """ Get stop lost variants """
        stop_lost = []
        stop_codon = FeatureLocation(start=stop_index, end=stop_index + 3)
        for variant in self.variants:
            if variant.variant.location.overlaps(stop_codon):
                stop_lost.append(variant.variant)
        return stop_lost

    def frames_shifted(self) -> int:
        """ Get total frames shifted of all variants """
        return sum(v.variant.frames_shifted() for v in self.variants) % 3

    def find_reference_next(self) -> PVGNode:
        """ Find and return the next reference node. The next reference node
        is defined as the out node that has not variant, or not any variant
        that is not in this current node. """
        if not self.out_nodes:
            return None
        this_variants = [x.variant for x in self.variants]
        for node in self.out_nodes:
            if not node.variants:
                return node
            that_variants = [x.variant for x in node.variants]
            if not any(it not in this_variants for it in that_variants):
                return node
        raise ValueError('None of the out nodes is reference.')

    def find_reference_prev(self) -> PVGNode:
        """ Find and return the previous reference node. The previous reference
        node is defined as the inbond node that has not variant, or any variant
        that is not in this current node. """
        if not self.in_nodes:
            return None
        if len(self.in_nodes) == 1:
            return list(self.in_nodes)[0]
        this_variants = [x.variant for x in self.variants]
        for node in self.in_nodes:
            if not node.variants:
                return node
            that_variants = [x.variant for x in node.variants]
            if not any(it not in this_variants for it in that_variants):
                return node
        raise ValueError('None of the in nodes is reference.')

    def find_least_varaint_prev(self) -> PVGNode:
        """ Find and return the inbonding node that has the least number of
        variants """
        if not self.in_nodes:
            return None
        if len(self.in_nodes) == 1:
            return list(self.in_nodes)[0]

        def sort_func(this:PVGNode, that:PVGNode):
            x = this.variants
            y = that.variants
            if len(x) < len(y):
                return -1
            if len(x) > len(y):
                return 1
            if len(x) == 0 and len(y) == 0:
                return -1
            for i,j in zip(x,y):
                if i.location < j.location:
                    return -1
                if i.location > j.location:
                    return 1
            return -1

        return sorted(self.in_nodes, key=cmp_to_key(sort_func))[0]

    def find_least_variant_next(self) -> PVGNode:
        """ Find the outbond node that has the least number of variants """
        if not self.out_nodes:
            return None
        if len(self.out_nodes) == 1:
            return list(self.out_nodes)[0]

        try:
            return self.find_reference_next()
        except ValueError:
            pass

        self_variants = {v.variant for v in self.variants}
        def sort_func(this:PVGNode, that:PVGNode):
            x = [v for v in this.variants if v.variant not in self_variants]
            y = [v for v in that.variants if v.variant not in self_variants]
            if len(x) < len(y):
                return -1
            if len(x) > len(y):
                return 1
            if len(x) == 0 and len(y) == 0:
                return -1
            for i,j in zip(x,y):
                if i.location < j.location:
                    return -1
                if i.location > j.location:
                    return 1
            return -1

        return sorted(self.out_nodes, key=cmp_to_key(sort_func))[0]

    def split_selenocysteines(self, index:int):
        """ Split selenocysteines """
        left_secs:List[seqvar.VariantRecordWithCoordinate] = []
        right_secs:List[seqvar.VariantRecordWithCoordinate] = []
        for sec in self.selenocysteines:
            if sec.location.start < index:
                left_secs.append(sec)
            else:
                right_secs.append(sec.shift(-index))
        return left_secs, right_secs

    def split_node(self, index:int, cleavage:bool=False, pre_cleave:bool=False,
            pop_collapse:bool=False, cleavage_range:Tuple[int,int]=None
            ) -> PVGNode:
        """ Split the sequence at the given position, and create a new node
        as the outbound edge. Variants will also be adjusted. For example:

        KAPGCFP -> split at position 3 -> KAP-GCFP

        The node with GCFP is returned.

        `cleavage_range` is a 2-tuple represents the range of the full cleavage
        pattern. Examples:
          - Peptide sequence of TTRTTT has cleavage site index of 2 and cleavage
            range of (2,4), the position of 'RT'. The `right_cleavage_pattern_start`
            of the left node is 2 and the `left_cleavage_pattern_end` of the right
            node is 1.
          - Peptide sequence of TTMRPTTT has cleavage site index of 3 and cleavage
            range of (2,5), the position of 'MRP'. The `right_cleavage_pattern_start`
            of the left node is 2 and the `left_cleavage_pattern_end` of the right
            node is 1.

        Args:
            index (int): the position to split
            cleavage (bool): If true, the cleavage attribute of the right node
                is update to be True.
            cleavage_range (Tuple[int, int]): Cleavage pattern rage.

        Returns:
            The new created node with the right part of the original sequence.
        """
        left_seq = self.seq[:index]
        right_seq = self.seq[index:]

        left_variants = []
        right_variants = []
        for variant in self.variants:
            if variant.location.start < index:
                if variant.location.end <= index:
                    left_variants.append(variant)
                else:
                    left_variants.append(variant[:index])
            elif variant.location.start == index and cleavage \
                    and not variant.downstream_cleavage_altering:
                cleave_alts = variant[index:index+1]
                cleave_alts = cleave_alts.shift(-1)
                cleave_alts.downstream_cleavage_altering = True
                left_variants.append(cleave_alts)

            if variant.location.end > index:
                right_variants.append(variant.shift(-index))
            elif variant.location.end == index and cleavage \
                    and not variant.upstream_cleavage_altering:
                cleave_alts = variant.shift(-index)
                cleave_alts.upstream_cleavage_altering = True
                right_variants.append(cleave_alts)

        self.seq = left_seq
        self.variants = left_variants

        new_node = PVGNode(
            seq=right_seq,
            reading_frame_index=self.reading_frame_index,
            variants=right_variants,
            orf=self.orf,
            was_bridge=self.was_bridge,
            pre_cleaved=pre_cleave,
            subgraph_id=self.subgraph_id,
            level=self.level,
            cpop_collapsed=self.cpop_collapsed,
            truncated=self.truncated,
            right_cleavage_pattern_start=self.right_cleavage_pattern_start
        )
        new_node.orf = self.orf

        left_secs, right_secs = self.split_selenocysteines(index)
        self.selenocysteines = left_secs
        new_node.selenocysteines = right_secs

        if cleavage:
            new_node.cleavage = True
            if cleavage_range:
                new_node.left_cleavage_pattern_end = cleavage_range[1] - index - 1

        if cleavage and cleavage_range:
            self.right_cleavage_pattern_start = cleavage_range[0]
        else:
            self.right_cleavage_pattern_start = None

        if pop_collapse:
            self.cpop_collapsed = True
            new_node.npop_collapsed = True
            if new_node.is_bridge():
                self.was_bridge = True
        else:
            self.cpop_collapsed = False

        self.truncated = False

        while self.out_nodes:
            node = self.out_nodes.pop()
            self.remove_out_edge(node)
            new_node.add_out_edge(node)
        self.add_out_edge(new_node)
        return new_node

    def truncate_right(self, i:int) -> PVGNode:
        """ Truncate the right i nucleotides off. """
        right_seq = self.seq[i:]
        left_variants = []
        right_variants = []

        for variant in self.variants:
            if variant.location.start < i:
                left_variants.append(variant)
            if variant.location.end > i:
                right_variants.append(variant.shift(-i))

        right_node = PVGNode(
            seq=right_seq,
            variants=right_variants,
            reading_frame_index=self.reading_frame_index,
            was_bridge=self.was_bridge,
            subgraph_id=self.subgraph_id,
            level=self.level,
            cpop_collapsed=self.cpop_collapsed,
            right_cleavage_pattern_start=self.right_cleavage_pattern_start
        )

        self.seq = self.seq[:i]
        self.variants = left_variants
        self.right_cleavage_pattern_start = None

        left_secs, right_secs = self.split_selenocysteines(i)
        self.selenocysteines = left_secs
        right_node.selenocysteines = right_secs

        return right_node

    def truncate_left(self, i:int) -> PVGNode:
        """ Truncate the left i nucleotides off. A new node with the left part
        of the sequences and variants associated is returned. The self node
        is updated with only the right part of the sequence and variants. """
        left_seq = self.seq[:i]
        left_variants = []
        right_variants = []

        for variant in self.variants:
            if variant.location.start < i:
                left_variants.append(variant)
            if variant.location.end > i:
                right_variants.append(variant.shift(-i))

        left_node = PVGNode(
            seq=left_seq,
            variants=left_variants,
            reading_frame_index=self.reading_frame_index,
            was_bridge=self.was_bridge,
            subgraph_id=self.subgraph_id,
            level=self.level,
            npop_collapsed=self.npop_collapsed,
            upstream_indel_map=self.upstream_indel_map,
            left_cleavage_pattern_end=self.left_cleavage_pattern_end
        )
        self.upstream_indel_map = {}

        self.seq = self.seq[i:]
        self.variants = right_variants
        self.left_cleavage_pattern_end = None

        left_secs, right_secs = self.split_selenocysteines(i)
        left_node.selenocysteines = left_secs
        self.selenocysteines = right_secs

        return left_node

    def append_left(self, other:PVGNode) -> None:
        """ Combine the other node to the left. """
        self.seq = other.seq + self.seq
        self.left_cleavage_pattern_end = other.left_cleavage_pattern_end
        variants = copy.copy(other.variants)
        for variant in self.variants:
            variants.append(variant.shift(len(other.seq.seq)))
        self.variants = variants
        self.upstream_indel_map = {k:copy.copy(v) for k,v in
            other.upstream_indel_map.items()}

        secs = copy.copy(other.selenocysteines)
        for sec in self.selenocysteines:
            sec = sec.shift(len(other.seq.seq))
            secs.append(sec)
        self.selenocysteines = secs
        self.left_cleavage_pattern_end = other.left_cleavage_pattern_end

    def append_right(self, other:PVGNode) -> None:
        """ Combine the other node the the right. """
        new_seq = self.seq + other.seq
        self.right_cleavage_pattern_start = other.right_cleavage_pattern_start
        variants = []
        for variant in self.variants:
            if not variant.downstream_cleavage_altering:
                variants.append(variant)
        self.variants = variants

        for variant in other.variants:
            self.variants.append(variant.shift(len(self.seq.seq)))

        for sec in other.selenocysteines:
            sec = sec.shift(len(self.seq.seq))
            self.selenocysteines.append(sec)

        self.seq = new_seq
        self.cpop_collapsed = other.cpop_collapsed
        self.truncated = other.truncated
        self.right_cleavage_pattern_start = other.right_cleavage_pattern_start

    def find_start_index(self) -> int:
        """ Find the start amino acid position """
        return self.seq.seq.find('M')

    def copy(self, in_nodes:bool=True, out_nodes:bool=True) -> PVGNode:
        """ Create a copy of the node """
        new_in_nodes = copy.copy(self.in_nodes) if in_nodes else set()
        new_out_nodes = copy.copy(self.out_nodes) if out_nodes else set()
        return PVGNode(
            seq=self.seq,
            variants=copy.copy(self.variants),
            in_nodes=new_in_nodes,
            out_nodes=new_out_nodes,
            cleavage=self.cleavage,
            truncated=self.truncated,
            orf=self.orf,
            selenocysteines=copy.copy(self.selenocysteines),
            reading_frame_index=self.reading_frame_index,
            was_bridge=self.was_bridge,
            subgraph_id=self.subgraph_id,
            level=self.level,
            npop_collapsed=self.npop_collapsed,
            cpop_collapsed=self.cpop_collapsed,
            upstream_indel_map={k:copy.copy(v) for k,v in self.upstream_indel_map.items()},
            left_cleavage_pattern_end=self.left_cleavage_pattern_end,
            right_cleavage_pattern_start=self.right_cleavage_pattern_start
        )

    def get_nearest_next_ref_index(self) -> int:
        """ get the nearest reference index """
        cur = self
        while not cur.seq.locations and cur.out_nodes:
            cur = cur.find_reference_next()

        if not cur.seq.locations:
            return -1
        return int(cur.seq.locations[0].ref.start)

    def get_orf_start(self, i:int=0) -> int:
        """ get the ORF start position at the reference transcript sequence
        given M is found at the query postion i.
        """
        k = self.reading_frame_index
        if self.seq.locations:
            for loc in self.seq.locations:
                if i < loc.query.start:
                    return loc.ref.start * 3 + k
                if loc.query.start <= i < loc.query.end:
                    return (i - loc.query.start + loc.ref.start) * 3 + k

        out_node = self.find_least_variant_next()
        if str(out_node.seq.seq) == '*' and not out_node.out_nodes:
            return -1
        if out_node.seq.locations:
            return out_node.seq.locations[0].ref.start * 3 + k
        # raise ValueError('Can not find ORF')
        return -1

    def get_subgraph_id_set(self) -> Set[str]:
        """ Get all subgraph ID as set """
        ids = set()
        for loc in self.seq.locations:
            if len(loc) == 0:
                continue
            ids.add(loc.ref.seqname)
        for v in self.variants:
            if v.variant.is_circ_rna():
                continue
            ids.add(v.location.seqname)
        return ids

    def get_subgraph_id_at(self, i:int) -> str:
        """ Get the subgraph ID at the given position """
        for loc in self.seq.locations:
            if i in loc.query:
                return loc.ref.seqname
            if loc.query.start > i:
                break
        for v in self.variants:
            if i in v.location:
                return v.location.seqname
            if v.location.start > i:
                break
        raise ValueError('Failed to find the subgraph ID.')

    def is_identical(self, other:PVGNode) -> bool:
        """ Checks if two nodes have the same sequence and same outbond node """
        return self.seq == other.seq and \
            self.out_nodes == other.out_nodes and \
            self.cleavage == other.cleavage and \
            self.reading_frame_index == other.reading_frame_index

    def is_less_mutated_than(self, other:PVGNode) -> bool:
        """ Checks if this node has less mutation than the other """
        if self.level != other.level:
            return self.level < other.level

        if len(self.variants) != len(other.variants):
            return len(self.variants) < len(other.variants)

        for x, y in zip(self.variants, other.variants):
            if x.location != y.location:
                return x.location < y.location

        for x, y in zip(self.variants, other.variants):
            # SNV is higher than RES
            if x.variant.type == 'SNV' and y.variant.type == 'RNAEditingSite':
                return True
            if x.variant.type == 'RNAEditingSite' and y.variant.type == 'SNV':
                return False
            # Otherwise don't really care about the order, just to make sure
            # the result is reproducible
            if x.variant.type != y.variant.type:
                return x.variant.type > y.variant.type

            if x.variant.alt != y.variant.type:
                return x.variant.alt > y.variant.type
        return True

    def transfer_in_nodes_to(self, other:PVGNode):
        """ transfer all in nodes of current node to the other """
        for in_node in copy.copy(self.in_nodes):
            if in_node in other.in_nodes:
                continue
            in_node.add_out_edge(other)

    def is_already_cleaved(self):
        """ Check if the node is already cleaved """
        return (self.cleavage and not self.pre_cleave) and \
            all(x.cleavage and not x.pre_cleave for x in self.out_nodes) and\
            all(all(y.cleavage and not y.pre_cleave for y in x.in_nodes)
                for x in self.out_nodes)

    def get_downstream_stop_altering_variants(self) -> Set[VariantRecord]:
        """ Get downstream stop altering variants """
        final_variants = set()
        for out_node in self.out_nodes:
            if len(out_node.seq.seq) == 1 and out_node.seq.seq.startswith('*'):
                stop_alts = set()
                stop_alts.update([x.variant for x in out_node.variants
                    if x.is_stop_altering and not x.downstream_cleavage_altering])
                stop_alts.update(out_node.upstream_indel_map.get(self, []))
                if not stop_alts:
                    return set()
                if not final_variants:
                    final_variants = stop_alts
        return final_variants

    def has_variant_at(self, start:int, end:int) -> bool:
        """ Checks if the node has any variant at a given position """
        loc = FeatureLocation(start=start, end=end)
        for variant in self.variants:
            if variant.location.overlaps(loc):
                return True
        return False

    def is_missing_variant(self, variant:seqvar.VariantRecord, upstream:PVGNode=None
            ) -> bool:
        """ Checks if in the coordinates of the node if any variant from a
        given set is missing. This is used for circRNA to ensure that the
        same variant is either on or off in each loop. """
        locs = []
        if self.seq.locations:
            for loc in self.seq.locations:
                tx_loc = FeatureLocation(
                    start=loc.get_ref_dna_start(),
                    end=loc.get_ref_dna_end()
                )
                locs.append(tx_loc)
        if self.variants:
            for v in self.variants:
                if v.variant.is_circ_rna():
                    continue
                locs.append(v.variant.location)
        variants = {x.variant for x in self.variants}
        if upstream:
            upstream_indels = self.upstream_indel_map.get(upstream, [])
            variants.update(v for v in upstream_indels)
        for loc in locs:
            if loc.overlaps(variant.location) \
                    and not any(variant == v for v in variants):
                return True
        return False

    def is_missing_any_variant(self, variants:Iterable[seqvar.VariantRecord],
            upstream:PVGNode=None) -> bool:
        """ Checks if the node is missing any of the variants. """
        return any(self.is_missing_variant(v, upstream) for v in variants)

    def has_variants_not_in(self, variants:Set[seqvar.VariantRecord]) -> bool:
        """ Checks if the node has any variants that is not in a given set. """
        if not variants:
            return False
        var_start = sorted(variants, key=lambda x: x.location.start)[0].location.start
        var_end = sorted(variants, key=lambda x: x.location.end)[-1].location.end
        var_range = FeatureLocation(start=var_start, end=var_end)
        for v in self.variants:
            if v.variant.is_circ_rna():
                continue
            if v.variant.location.overlaps(var_range) \
                    and v.variant not in variants:
                return True
        return False

    def is_hybrid_node(self, tree:SubgraphTree, check_hard:bool=True) -> bool:
        """ Checks if the node is hybrid. A hybrid node is when two parts of
        the node sequence are from different subgraphs, and have the different
        states. """
        variants = {v.variant for v in self.variants if not v.variant.is_circ_rna()}

        seq_iter = iter(self.seq.locations)
        var_iter = iter(self.variants)

        cur_seq = next(seq_iter, None)
        cur_var = next(var_iter, None)

        start, end = 0, 0
        start_hard, end_hard = 0, 0
        cur_level = -1

        def set_hard_start_and_end(loc, start, end):
            start_hard = start
            end_hard = end
            if check_hard:
                if loc.start_offset != 0:
                    start_hard += 1
                if loc.end_offset != 0:
                    end_hard -= 1
            return start_hard, end_hard

        while cur_seq or cur_var:
            if cur_var and cur_var.variant.is_circ_rna():
                cur_var = next(var_iter, None)
                continue

            seq_level, var_level = -1, -1
            seq_start, seq_end = -1, -1
            var_start, var_end = -1, -1

            if cur_seq:
                seq_level = tree[cur_seq.ref.seqname].level
                seq_start = cur_seq.query.start
                seq_end = cur_seq.query.end

            if cur_var:
                var_level = tree[cur_var.location.seqname].level
                var_start = cur_var.location.start
                var_end = cur_var.location.end

            seq_is_smaller = (not cur_var ) \
                or (cur_seq and (seq_level < var_level or seq_start < var_start))

            if seq_is_smaller:
                if cur_level == -1:
                    cur_level = seq_level
                    start, end = seq_start, seq_end
                    start_hard, end_hard = set_hard_start_and_end(cur_seq.query, start, end)

                elif seq_level > cur_level:
                    seg = self[start:end]
                    if seg.is_missing_any_variant(variants):
                        return True
                    if start_hard < end_hard:
                        seg = self[start_hard:end_hard]
                        seg.variants = [v for v in seg.variants
                            if not v.variant.is_circ_rna()
                            and tree[v.location.seqname].level == cur_level]
                        if seg.is_missing_any_variant(variants):
                            return True
                    cur_level = seq_level
                    start, end = seq_start, seq_end
                    start_hard, end_hard = set_hard_start_and_end(cur_seq.query, start, end)
                else:
                    end = seq_end
                    start_hard, end_hard = set_hard_start_and_end(cur_seq.query, start, end)

                if cur_seq is self.seq.locations[-1]:
                    if cur_var:
                        cur_var = next(var_iter, None)
                        continue

                    end = seq_end
                    seg = self[start:end]
                    if seg.is_missing_any_variant(variants):
                        return True
                    if start_hard < end_hard:
                        seg = self[start_hard:end_hard]
                        if seg.is_missing_any_variant(variants):
                            return True

                cur_seq = next(seq_iter, None)

            else:
                if cur_level == -1:
                    cur_level = var_level
                    start, end = var_start, var_end
                    start_hard, end_hard = set_hard_start_and_end(cur_var.location, start, end)
                elif var_level > cur_level:
                    seg = self[start:end]
                    if seg.is_missing_any_variant(variants):
                        return True
                    if start_hard < end_hard:
                        seg = self[start_hard:end_hard]
                        seg.variants = [v for v in seg.variants
                            if not v.variant.is_circ_rna()
                            and tree[v.location.seqname].level == cur_level]
                        if seg.is_missing_any_variant(variants):
                            return True
                    cur_level = var_level
                    start, end = var_start, var_end
                    start_hard, end_hard = set_hard_start_and_end(cur_var.location, start, end)
                else:
                    end = var_end
                    start_hard, end_hard = set_hard_start_and_end(cur_var.location, start, end)

                if cur_var is self.variants[-1] and not cur_seq:
                    if cur_seq:
                        cur_seq = next(seq_iter, None)
                        continue

                    end = var_end
                    seg = self[start:end]
                    if seg.is_missing_any_variant(variants):
                        return True
                    if start_hard < end_hard:
                        seg = self[start_hard:end_hard]
                        if seg.is_missing_any_variant(variants):
                            return True

                cur_var = next(var_iter, None)

        return False

    def any_unaccounted_downstream_cleavage_or_stop_altering(self,
            variants:Iterable[seqvar.VariantRecord]) -> bool:
        """ Checks if the node has any unaccounted cleavage or stop altering
        mutation from a given list. The first amino acid from the downstream
        node and the last amino acid from the upstream one will be checked if
        they contain any variant in the list of variant provided and not having
        any variant that are not in the list.
        """
        variants = {v for v in variants if not v.is_circ_rna()}

        boundary_nodes:List[PVGNode] = []

        # stop and downstream cleavage gain
        if len(self.get_out_nodes()) == 1:
            downstream = self.get_out_nodes()[0]
            boundary_node = downstream[:1]
            b_vars = []
            for v in boundary_node.variants:
                if not v.upstream_cleavage_altering and not v.variant.is_circ_rna():
                    b_vars.append(v)
            boundary_node.variants = b_vars
            if not (downstream.seq.seq == '*' and not downstream.get_out_nodes()):
                boundary_nodes.append(boundary_node)

        # upstream cleavage gain cleavage gain
        if len(self.get_in_nodes()) == 1:
            upstream = self.get_in_nodes()[0]
            boundary_node = upstream[-1:]
            b_vars = []
            for v in boundary_node.variants:
                if not v.downstream_cleavage_altering and not v.variant.is_circ_rna():
                    b_vars.append(v)
            boundary_node.variants = b_vars
            if upstream.seq is not None:
                boundary_nodes.append(boundary_node)

        for node in boundary_nodes:
            if node.is_missing_any_variant(variants, self):
                return True
            if node.has_variants_not_in(variants):
                return True
        return False

    def fix_selenocysteines(self,
            sect_variants:List[seqvar.VariantRecordWithCoordinate],
            subgraphs:SubgraphTree) -> None:
        """ Fix Selenocysteines from the amin acid sequence. """
        iter_loc = iter(self.seq.locations)
        iter_sec = iter(sect_variants)
        loc = next(iter_loc, None)
        sect = next(iter_sec, None)

        sects:List[seqvar.VariantRecordWithCoordinate] = []

        while loc and sect:
            if loc.ref.seqname not in subgraphs.data \
                    or subgraphs[loc.ref.seqname].level != 0:
                loc = next(iter_loc, None)
                continue

            rf_index = loc.query.reading_frame_index
            if rf_index != sect.location.start % 3:
                if loc.get_ref_dna_start() > sect.location.start:
                    sect = next(iter_sec, None)
                else:
                    loc = next(iter_loc, None)
                continue

            dna_start = loc.get_ref_codon_start()
            dna_end = loc.get_ref_codon_end()

            if loc.query.start_offset > 0:
                ref_codon_start = dna_start + 3
            else:
                ref_codon_start = dna_start

            if loc.query.end_offset > 0:
                ref_codon_end = dna_end - 3
            else:
                ref_codon_end = dna_end

            if ref_codon_start >= ref_codon_end:
                loc = next(iter_loc, None)
                continue

            dna_loc = FeatureLocation(ref_codon_start, ref_codon_end)
            if dna_loc.is_superset(sect.location):
                k = loc.query.start + int((sect.location.start - dna_start) / 3)
                sect_local = seqvar.VariantRecordWithCoordinate(
                    location=FeatureLocation(k, k+1, seqname=sect.variant.transcript_id),
                    variant=sect.variant
                )
                if self.seq.seq[k] != '*':
                    get_logger().warning(
                        'The codon at the given Selenocysteine position is not a stop codon.'
                    )
                sects.append(sect_local)

                sect = next(iter_sec, None)
                continue

            if dna_loc > sect.location:
                sect = next(iter_sec, None)
            else:
                loc = next(iter_loc, None)

        if not sects:
            return

        for i,sect in enumerate(sects):
            k = sect.location.start
            if i == 0:
                new_seq = self.seq.seq[:k]
            else:
                new_seq += self.seq.seq[sects[i-1].location.start + 1:k]
            new_seq += 'U'
            if sect is sects[-1]:
                if k + 1 < len(self.seq.seq):
                    new_seq += self.seq.seq[k+1:]

        self.selenocysteines = sects
        self.seq.seq = new_seq



""" Model for protein or peptide sequence data
"""
from __future__ import annotations
from typing import Set, TYPE_CHECKING
from Bio import SeqIO
from moPepGen.aa.AminoAcidSeqRecord import AminoAcidSeqRecord
from moPepGen.version import MetaVersion


# To avoid circular import
if TYPE_CHECKING:
    from moPepGen.gtf import GenomicAnnotation

class AminoAcidSeqDict(dict):
    """ AminoAcidSeqDict """
    def __init__(self, *args, **kwargs):
        """ Constructor """
        for val in kwargs.values():
            if not isinstance(val, AminoAcidSeqRecord):
                raise ValueError(
                    'The value of a DNASeqDict must be AminoAcidSeqRecord.'
                )
        super().__init__(*args, **kwargs)
        self.version = MetaVersion()

    def __setitem__(self, k:str, v:AminoAcidSeqRecord)->None:
        """ set item """
        if not isinstance(v, AminoAcidSeqRecord):
            raise ValueError('The value of a AASeqDict must be '
            'AminoAcidSeqRecord.')
        super().__setitem__(k, v)

    def dump_fasta(self, path:str, source:str=None)->None:
        """ Dump a FASTA file into an AminoAcidSeqDict

        Args:
            path (str): Path to the FASTA file of protein sequences.
            source (str): The source of the FASTA file. Can be 'GENCODE' or
                'ENSEMBL'. If None is given, it will be infered by trying
                the first 100 records. Default to None.
        """
        count = 0
        if not source:
            inferred = set()
        record:AminoAcidSeqRecord
        for record in SeqIO.parse(path, 'fasta'):
            record.__class__ = AminoAcidSeqRecord
            if count > 100 and not source:
                source = inferred.pop()

            if not source:
                count += 1
                inferred.add(record.infer_ids(style=source))
            else:
                record.infer_ids(source)
            if record.transcript_id in self.keys():
                raise ValueError(
                    'Duplicated seqnames found in FASTA file: ' + path
                )
            self[record.transcript_id] = record

    def create_unique_peptide_pool(self, anno:GenomicAnnotation,
            rule:str, exception:str=None, miscleavage:int=2, min_mw:float=500.,
            min_length:int=7, max_length:int=25)->Set[str]:
        """ Create a unique piptide pool.

        All peptides with I (isoleucine)
        replaced with L (leucine) are also added to the pool. This is used to
        filter variant peptides. For proteins that starts with X, the part of
        the sequence before the first cleave site is removed. Protein sequences
        that starts with X usually means that the 5' CDS is incomplete.

        For a given gene, if the cds start is known (i.e. the gene model does not
        have the cds_start_NF tag from the GTF file), the peptide with and
        without the leading methionine will both be included. This is because
        sometimes the start codon methionine is cleaved spontaneously inside the
        cell.

        Args:
            anno (GenomicAnnotation): Genomic annotation parsed from GTF.
            rule (str): The rule for enzymatic cleavage, e.g., trypsin.
            exception (str): The exception for cleavage rule.
            start (int): Index to start searching.
            min_mw (float): Minimal molecular weight of the peptides to report.
                Defaults to 500.
            min_length (int): Minimal length of the peptides to report, inclusive.
                Defaults to 7.
            max_length (int): Maximum length of the peptides to report, inclusive.
                Defaults to 25.

        Returns:
            A set of unique peptides as string.
        """
        pool = set()
        protein: AminoAcidSeqRecord
        it = iter(self.values())
        protein = next(it, None)
        while protein:
            tx_id = protein.transcript_id
            if tx_id in anno.transcripts:
                cds_start_nf = anno.transcripts[tx_id].is_cds_start_nf()
            else:
                cds_start_nf = False
            if protein.seq.startswith('X'):
                protein.seq = protein.seq.lstrip('X')

            stop_site = protein.seq.find('*')
            if stop_site > -1:
                protein = protein[:stop_site]

            try:
                peptides = protein.enzymatic_cleave(
                    rule=rule,
                    exception=exception,
                    miscleavage=miscleavage,
                    min_mw=min_mw,
                    min_length=min_length,
                    max_length=max_length,
                    cds_start_nf=cds_start_nf
                )
            except ValueError as e:
                msg = "'X' is not a valid unambiguous letter for protein"
                if e.args[0] == msg:
                    protein.seq = protein.seq.split('X')[0]
                    continue
                raise e
            for peptide in peptides:
                pool.add(str(peptide.seq))
                # Convert all I with L and add to the canonical peptide pool.
                pool.add(str(peptide.seq).replace('I', 'L'))

            protein = next(it, None)

        return pool



""" Variant Record Pool """
from __future__ import annotations
import copy
from typing import Dict, IO, Iterable, List, TYPE_CHECKING, Union
from pathlib import Path
from moPepGen import ERROR_INDEX_IN_INTRON, check_sha512, circ, constant
from moPepGen.seqvar.GVFIndex import GVFPointer, iterate_pointer
from moPepGen.seqvar.GVFMetadata import GVFMetadata
from . import VariantRecord


# To avoid circular import
if TYPE_CHECKING:
    from moPepGen.gtf import GenomicAnnotation
    from moPepGen.dna import DNASeqDict, DNASeqRecordWithCoordinates


class TranscriptionalVariantSeries():
    """ Variants associated with a particular transcript """
    def __init__(self, transcriptional:List[VariantRecord]=None,
            intronic:List[VariantRecord]=None, fusion:List[VariantRecord]=None,
            circ_rna:List[circ.CircRNAModel]=None
            ) -> None:
        """ constructor """
        self.transcriptional = transcriptional or []
        self.intronic = intronic or []
        self.fusion = fusion or []
        self.circ_rna = circ_rna or []

    def __copy__(self) -> TranscriptionalVariantSeries:
        """ copy """
        return self.__class__(
            transcriptional=copy.copy(self.transcriptional),
            intronic=copy.copy(self.intronic),
            fusion=copy.copy(self.fusion),
            circ_rna=copy.copy(self.circ_rna)
        )

    def sort(self):
        """ Sort each slot of variants in order """
        self.transcriptional.sort()
        self.intronic.sort()
        self.fusion.sort()

    def get_additional_transcripts(self) -> List[str]:
        """ Get any additioanl transcript IDs that are associated with any
        variants, for example, fusion. """
        transcripts = []
        for variant in self.fusion:
            transcripts.append(variant.attrs['ACCEPTER_TRANSCRIPT_ID'])
        return transcripts

    def is_gene_sequence_needed(self) -> bool:
        """ Check if the gene sequence is needed when calling variant peptides.
        Usually for variants that invole retaining of a complete or partial
        intron. """
        return len(self.fusion) > 0 or len(self.circ_rna) > 0 or \
            any(x.type in ['Insertion', 'Substitution'] for x in self.transcriptional)

    def is_empty(self) -> bool:
        """ check if the series is empty """
        return len(self.transcriptional) == 0 and len(self.fusion) == 0 and \
            len(self.circ_rna) == 0

    def has_any_noncanonical_transcripts(self) -> bool:
        """ check if the series has any noncanonical transcripts """
        return len(self.fusion) > 0 or len(self.circ_rna) > 0 or \
            any(x.type in constant.ALTERNATIVE_SPLICING_TYPES for x in self.transcriptional)

    def has_any_alternative_splicing(self) -> bool:
        """ Check if there is any alternative splicing """
        return any(x.type in constant.ALTERNATIVE_SPLICING_TYPES for x in self.transcriptional)

    def get_highest_hypermutated_region_complexity(self, distance:int=6):
        """ Calculate the number of variants in the most hypermutated region """
        end = 0
        max_n = 0
        cur_n = 0
        for variant in self.transcriptional:
            if variant.type in constant.ALTERNATIVE_SPLICING_TYPES:
                continue
            if variant.location.start - end >= distance:
                end = variant.location.end
                max_n = max(max_n, cur_n)
                cur_n = 1
                continue
            cur_n += 1
            end = variant.location.end
        return max(max_n, cur_n)

class VariantRecordPoolOnDisk():
    """ Variant record pool in disk """
    def __init__(self, pointers:Dict[str, List[GVFPointer]]=None,
            gvf_files:List[Path]=None, gvf_handles:List[IO]=None,
            anno:GenomicAnnotation=None, genome:DNASeqDict=None) -> None:
        """ constructor """
        self.pointers = pointers or {}
        self.gvf_files = gvf_files or []
        self.gvf_handles = gvf_handles or []
        self.anno = anno
        self.genome = genome

    def __contains__(self, key:str):
        """ conteins """
        return key in self.pointers

    def __iter__(self) -> Iterable[str]:
        """ generator """
        for key in self.pointers:
            yield key

    def __getitem__(self, key:str) -> TranscriptionalVariantSeries:
        """ Load variants and return as a TranscriptVariants object """
        records:List[Union[VariantRecord, circ.CircRNAModel]] = []
        for pointer in self.pointers[key]:
            records += pointer.load()
        records = set(records)
        series = TranscriptionalVariantSeries()
        cached_seqs:Dict[str, DNASeqRecordWithCoordinates] = {}
        for record in records:
            if isinstance(record, circ.CircRNAModel):
                series.circ_rna.append(record)
                continue

            tx_id = record.transcript_id

            if record.is_fusion():
                record.shift_breakpoint_to_closest_exon(self.anno)
                tx_record = record.to_transcript_variant(
                    self.anno, self.genome, tx_id, cached_seqs
                )
                series.fusion.append(tx_record)
                continue

            if record.is_spanning_over_splicing_site(self.anno, tx_id):
                continue

            try:
                tx_record = record.to_transcript_variant(
                    self.anno, self.genome, tx_id
                )
                if tx_record.type == 'Deletion':
                    if tx_id in cached_seqs:
                        tx_seq = cached_seqs[tx_id]
                    else:
                        tx_model = self.anno.transcripts[tx_id]
                        chrom = tx_model.transcript.chrom
                        tx_seq = tx_model.get_transcript_sequence(self.genome[chrom])
                    tx_record.shift_deletion_up(tx_seq)
                series.transcriptional.append(tx_record)
            # except err.FusionBreakpointIsEndOfTranscript as e:
            #     continue
            except ValueError as e:
                if e.args[0] == ERROR_INDEX_IN_INTRON:
                    series.intronic.append(record)
                else:
                    raise e

        series.sort()
        return series

    def load_index(self, index_file:Path, gvf_file:Path, gvf_handle:IO):
        """ Load index file """
        with open(gvf_file, 'rt') as handle:
            metadata = GVFMetadata.parse(handle)
        is_circ_rna = metadata.is_circ_rna()
        gvf_handle.seek(0)
        with open(index_file, 'rt') as idx_handle:
            reader = GVFPointer.parse(
                index_handle=idx_handle, gvf_handle=gvf_handle,
                is_circ_rna=is_circ_rna
            )
            for pointer in reader:
                if pointer.key in self.pointers:
                    self.pointers[pointer.key].append(pointer)
                else:
                    self.pointers[pointer.key] = [pointer]

    def generate_index(self, gvf_file:Path, gvf_handle:IO):
        """ generate index """
        with open(gvf_file, 'rt') as handle:
            metadata = GVFMetadata.parse(handle)
        is_circ_rna = metadata.is_circ_rna()
        gvf_handle.seek(0)
        for pointer in iterate_pointer(gvf_handle, is_circ_rna):
            if pointer.key in self.pointers:
                self.pointers[pointer.key].append(pointer)
            else:
                self.pointers[pointer.key] = [pointer]
        gvf_handle.seek(0)

    @staticmethod
    def validate_gvf_index(gvf_file:Path, idx_file:Path):
        """ Validate the GVF file with index """
        sum_expect = None
        with open(gvf_file, 'rb') as gvf_handle:
            sum_actual = check_sha512(gvf_handle)

        with open(idx_file, 'rt') as idx_handle:
            for line in idx_handle:
                line:str
                if line.startswith('#'):
                    line = line.rstrip().lstrip('# ')
                    if line.startswith('CHECKSUM='):
                        sum_expect = line.split('=')[1]

                        break
                else:
                    break

        if sum_expect is None:
            raise ValueError('Cannot find checksum value from the idx file.')
        if sum_actual == sum_expect:
            return True

        raise ValueError("GVF checksum don't match.")

    def filter_variants(self, gene_id:str=None, tx_ids:List[str]=None,
            exclude_type:List[str]=None, start:int=None, end:int=None,
            intron:bool=True, segments:Iterable[VariantRecord]=None,
            return_coord:str='gene') -> List[VariantRecord]:
        """ filter variants """
        if return_coord not in ['gene', 'transcript']:
            raise ValueError(
                "Don't know how to return variants in coordinate of "
                f"{return_coord}. return_coord must be either 'gene' or "
                "'transcript'"
            )
        if return_coord == 'transcript' and intron is True:
            raise ValueError(
                "Don't know how to return intronic variants in transcript"
                "coordinates."
            )
        if segments:
            def _filter(x):
                for segment in segments:
                    if segment.location.is_superset(x.location):
                        return True
                return False
        elif start is not None and end is not None:
            _filter = lambda x: x.location.start > start and x.location.end < end
        elif start is not None:
            _filter = lambda x: x.location.start > start
        elif end is not None:
            _filter = lambda x: x.location.end < end
        else:
            raise ValueError('Arguments provided do not match requirement.')

        records = set()
        if tx_ids:
            _tx_ids = set(tx_ids)
        else:
            _tx_ids = set()
        if gene_id:
            _tx_ids.update(self.anno.genes[gene_id].transcripts)

        for tx_id in _tx_ids:
            gene_id = self.anno.transcripts[tx_id].transcript.gene_id
            if tx_id not in self.pointers:
                continue
            series = self[tx_id]
            for record in series.transcriptional:
                if record.type in exclude_type:
                    continue
                record_gene = self.anno.variant_coordinates_to_gene(record, gene_id)
                if _filter(record_gene):
                    if return_coord == 'gene':
                        records.add(record_gene)
                    else:
                        records.add(record)

            if not intron:
                continue

            for record in series.intronic:
                if record.type in exclude_type:
                    continue
                if _filter(record):
                    if return_coord == 'gene':
                        records.add(record)

        records = list(records)
        records.sort()
        return records

    def get_transcript_order(self) -> List[str]:
        """ Get the order of transcripts to call variant peptides """
        def sort_key(key:str):
            return self[key].get_highest_hypermutated_region_complexity()
        return sorted(self.pointers.keys(), key=sort_key, reverse=True)

class VariantRecordPoolOnDiskOpener():
    """ Helper class to open all GVF files of a VariantRecordPoolOnDisk. """
    def __init__(self, pool:VariantRecordPoolOnDisk):
        """ Constructor """
        self.pool = pool

    def __enter__(self):
        """ enter """
        self.open()
        return self.pool

    def __exit__(self, exception_type, exception_value, exception_traceback):
        """ exit """
        self.close()

    def open(self):
        """ Open all GVF files """
        for file in self.pool.gvf_files:
            gvf_handle = file.open('rb')
            self.pool.gvf_handles.append(gvf_handle)
            idx_path = file.with_suffix(file.suffix + '.idx')
            if idx_path.exists():
                self.pool.validate_gvf_index(file, idx_path)
                self.pool.load_index(idx_path, file, gvf_handle)
            else:
                self.pool.generate_index(file, gvf_handle)

    def close(self):
        """ Close all file handles """
        for handle in self.pool.gvf_handles:
            handle.close()
